# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import urllib
from deejayd.ui import log

def quote_uri(path):
    if type(path) is unicode:
        path = path.encode('utf-8')
    return "file://%s" % urllib.quote(path)

def str_encode(data, charset = 'utf-8', errors='strict'):
    if type(data) is unicode: return data
    try: rs = data.decode(charset, errors)
    except UnicodeError:
        log.err(_("%s string has wrong characters, skip it") %\
          data.decode(charset, "ignore").encode("utf-8","ignore"))
        raise UnicodeError
    return unicode(rs)

def format_time(time):
    """Turn a time value in seconds into hh:mm:ss or mm:ss."""
    if time >= 3600: # 1 hour
        # time, in hours:minutes:seconds
        return "%d:%02d:%02d" % (time // 3600, (time % 3600) // 60, time % 60)
    else:
        # time, in minutes:seconds
        return "%d:%02d" % (time // 60, time % 60)

def format_time_long(time):
    """Turn a time value in seconds into x hours, x minutes, etc."""
    if time < 1: return _("No time information")
    cutoffs = [
        (60, "%d seconds", "%d second"),
        (60, "%d minutes", "%d minute"),
        (24, "%d hours", "%d hour"),
        (365, "%d days", "%d day"),
        (None, "%d years", "%d year"),
    ]
    time_str = []
    for divisor, plural, single in cutoffs:
        if time < 1: break
        if divisor is None: time, unit = 0, time
        else: time, unit = divmod(time, divisor)
        if unit: time_str.append(ngettext(single, plural, unit) % unit)
    time_str.reverse()
    if len(time_str) > 2: time_str.pop()
    return ", ".join(time_str)

    ngettext("%d second", "%d seconds", 1)
    ngettext("%d minute", "%d minutes", 1)
    ngettext("%d hour", "%d hours", 1)
    ngettext("%d day", "%d days", 1)
    ngettext("%d year", "%d years", 1)

def get_playlist_file_lines(URL):
    pls_handle = urllib.urlopen(URL)
    playlist = pls_handle.read()

    return playlist.splitlines()

def get_uris_from_pls(URL):
    uris = []
    lines = get_playlist_file_lines(URL)
    for line in lines:
        if line.lower().startswith("file") and line.find("=")!=-1:
            uris.append(line[line.find("=")+1:].strip())

    return uris

def get_uris_from_m3u(URL):
    uris = []
    lines = get_playlist_file_lines(URL)
    for line in lines:
        if not line.startswith("#") and line.strip()!="":
            uris.append(line.strip())

    return uris

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


class _DBQuery(object):

    def __init__(self, table_name):
        self.table_name = table_name

    def get_args(self):
        raise NotImplementedError

    def to_sql(self):
        return str(self)


class SimpleSelect(_DBQuery):

    def __init__(self, table_name):
        super(SimpleSelect, self).__init__(table_name)
        self.selects = []
        self.orders = []
        self.wheres, self.wheres_args = [], []

    def select_column(self, *__args, **__kw):
        for col in __args:
            self.selects.append("%s.%s" % (self.table_name, col))

    def order_by(self, column, desc = False):
        self.orders.append("%s.%s" % (self.table_name, column))

    def append_where(self, where_query, args):
        self.wheres.append(where_query)
        self.wheres_args.extend(args)

    def get_args(self):
        return self.wheres_args

    def __str__(self):
        return "SELECT DISTINCT %s FROM %s WHERE %s"\
               % ( ', '.join(self.selects),
                   self.table_name,
                   ' AND '.join(self.wheres) or 1,
                 )

class MediaSelectQuery(SimpleSelect):

    def __init__(self):
        super(MediaSelectQuery, self).__init__('library')
        self.joins = []
        self.limit = None
        self.__joined_tags = []
        self.id = False

    def select_id(self):
        self.id = True

    def select_column(self, column_name, table_name=None):
        if not table_name:
            table_name = self.table_name
        self.selects.append("%s.%s" % (table_name, column_name))

    def select_tag(self, tagname):
        self.select_column('value', tagname)
        self.join_on_tag(tagname)

    def order_by_tag(self, tagname, desc = False):
        order = "%s.value" % tagname
        if desc: order = "%s DESC" % order
        self.orders.append(order)
        self.join_on_tag(tagname)

    def join_on_tag(self, tagname):
        if tagname not in self.__joined_tags:
            self.__joined_tags.append(tagname)
            j_st = "JOIN media_info %(tag)s ON %(tag)s.id = library.id\
                                            AND %(tag)s.ikey = '%(tag)s'"\
                   % { 'tag' : tagname }
            self.joins.append(j_st)

    def set_limit(self, limit):
        self.limit = limit

    def __str__(self):
        orders, limit = None, None
        if len(self.orders) >= 1:
            orders = 'ORDER BY ' + ', '.join(self.orders)
        if self.limit is not None:
            limit = "LIMIT %s" % str(self.limit)


        return "SELECT DISTINCT %s %s FROM %s %s WHERE %s %s %s"\
               % (self.id and 'library.id,' or '',
                  ', '.join(self.selects),
                  self.table_name,
                  ' '.join(self.joins),
                  ' AND '.join(self.wheres) or 1,
                  orders or '',
                  limit or '')


class EditRecordQuery(_DBQuery):

    def __init__(self, table_name):
        super(EditRecordQuery, self).__init__(table_name)
        self.dbvalues = {}
        self.update_id = None

    def add_value(self, column_name, column_value):
        self.dbvalues[column_name] = column_value

    def set_update_id(self, key, id):
        self.update_key = key
        self.update_id = id

    def get_args(self):
        args = self.dbvalues.values()
        if self.update_id:
            args.append(self.update_id)
        return args

    def __str__(self):
        if self.update_id:
            sets_st = ["%s = %%s" % x for x in self.dbvalues.keys()]
            query = "UPDATE %s SET %s WHERE %s"\
                    % ( self.table_name,
                        ', '.join(sets_st),
                        "%s.%s = %%s" % (self.table_name, self.update_key),
                      )
        else:
            query = "INSERT INTO %s(%s) VALUES(%s)"\
                    % ( self.table_name,
                        ', '.join(self.dbvalues.keys()),
                        ', '.join(["%s" for x in self.get_args()]),
                      )
        return query


def query_decorator(answer_type):
    def query_decorator_instance(func):

        def query_func(self, *__args, **__kw):
            cursor = self.connection.cursor()
            rs = func(self, cursor, *__args, **__kw)
            if answer_type == "lastid":
                rs = self.connection.get_last_insert_id(cursor)
            elif answer_type == "rowcount":
                rs = cursor.rowcount
            elif answer_type == "fetchall":
                rs = list(cursor.fetchall())
            elif answer_type == "fetchone":
                rs = cursor.fetchone()
            elif answer_type == "medialist":
                rs = self._medialist_answer(cursor.fetchall(),__kw['infos'])

            cursor.close()
            return rs

        return query_func

    return query_decorator_instance


# vim: ts=4 sw=4 expandtab



#!/usr/bin/env python

# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Use to create documentation of the protocol
"""

# init translation
import gettext
from deejayd.ui.i18n import DeejaydTranslations
try: t = gettext.translation("deejayd", class_=DeejaydTranslations)
except IOError:
    t = DeejaydTranslations()
t.install()

from twisted.python import reflect
from deejayd.mediafilters import *
from deejayd.interfaces import DeejaydSignal
from deejayd.rpc import protocol
from deejayd.rpc.jsonbuilders import JSONRPCResponse, JSONRPCRequest,\
                                     Get_json_filter, DeejaydJSONSignal

common_request = [
        {"prefix": "", "desc": "General Commands",\
                       "object": protocol._DeejaydMainJSONRPC},
        {"prefix": "player.", "desc": "Player Commands",\
                       "object": protocol.DeejaydPlayerJSONRPC},
        {"prefix": "audiolib.", "desc": "Audio Library Commands",\
                       "object": protocol.DeejaydAudioLibraryJSONRPC},
        {"prefix": "videolib.", "desc": "Video Library Commands",\
                       "object": protocol.DeejaydVideoLibraryJSONRPC},
        {"prefix": "playlist.", "desc": "Playlist Mode Commands",\
                       "object": protocol.DeejaydPlaylistModeJSONRPC},
        {"prefix": "panel.", "desc": "Panel Mode Commands",\
                       "object": protocol.DeejaydPanelModeJSONRPC},
        {"prefix": "video.", "desc": "Video Mode Commands",\
                       "object": protocol.DeejaydVideoModeJSONRPC},
        {"prefix": "webradio.", "desc": "Webradio Mode Commands",\
                       "object": protocol.DeejaydWebradioModeJSONRPC},
        {"prefix": "dvd.", "desc": "Dvd Mode Commands",\
                       "object": protocol.DeejaydDvdModeJSONRPC},
        {"prefix": "queue.", "desc": "Queue Commands",\
                       "object": protocol.DeejaydQueueJSONRPC},
        {"prefix": "queue.", "desc": "Queue Commands",\
                       "object": protocol.DeejaydQueueJSONRPC},
        {"prefix": "recpls.", "desc": "Recorded Playlist Commands",\
                       "object": protocol.DeejaydRecordedPlaylistJSONRPC},
    ]

class WikiFormat:

    def commandDoc(self):
        return """
As written in specification, request is like that :
{{{
`%(request)s`
}}}

""" % {
    "request": JSONRPCRequest("method_name",\
                          ["params1", "params2"], id="id").to_pretty_json(),\
    }

    def answerDoc(self):
        return """
As written in specification, response is like that :
{{{
%s
}}}

For deejayd, result parameter has always the same syntax :
{{{
`{
    "type": answer_type,
    "answer": the real answer value
}`
}}}
With response types equals to:
    * ack
    * list
    * dict
    * mediaList
    * dvdInfo
    * fileAndDirList
""" % JSONRPCResponse("deejayd_response", "id").to_pretty_json()

    def formatSectionDoc(self, section):
        cmds = reflect.prefixedMethodNames(section["object"], 'jsonrpc_')
        cmds = [getattr(section["object"], 'jsonrpc_%s'%cmd) for cmd in cmds]
        return """
=== `%(section)s` ===

%(commands)s
""" % {
        "section": section["desc"],
        "commands": "\n\n".join(map(self.formatCommandDoc, cmds,\
                        [section["prefix"] for i in range(len(cmds))])),
    }

    def formatCommandDoc(self, cmd, prefix = ""):
        args = ''

        command_args = cmd.params or []
        for arg in command_args:
            props = []

            # An argument is optional by default
            if 'req' not in arg.keys():
                arg['req'] = False
            if arg['req']:
                props.append('Mandatory')
            else:
                props.append('Optional')

            args += "  * {{{%(name)s}}} (%(props)s) : %(type)s\n"\
                        % { 'name':  arg['name'],
                            'props': ' and '.join(props),
                            'type' : arg['type'] }

        if len(command_args) == 0:
            args = "  * ''This command does not accept any argument.''\n"

        rvalues = None
        try:
            if isinstance(cmd.answer_type, list):
                rvalues = cmd.answer_type
            else:
                rvalues = [cmd.answer_type]
        except AttributeError:
            rvalues = ['ack']

        return """==== `%(name)s` ====

%(desc)s

Arguments :
%(args)s
Expected return value : ''`%(rvalues)s`''

""" % { 'name'    : prefix+cmd.__name__[8:],
        'desc'    : cmd.__doc__.strip('\n'),
        'args'    : args,
        'rvalues' : rvalues }

    def build(self, sections):
        filter = And(Equals("artist", "artist_name"),\
                Or(Contains("genre", "Rock"), Higher("Rating", "4")))
        signal = DeejaydSignal("signal_name", {"attr1": "value1"})
        return """= deejayd - JSON-RPC Protocol =

Deejayd protocol follows JSON-RPC 1.0 specification available
[http://json-rpc.org/wiki/specification here].
All data between the client and server is encoded in UTF-8.

== Commands Format ==

%(cmd_format)s

== Response Format ==

%(answer)s

== Specific Objects ==

=== Mediafilter Objects ===

Mediafilter object has been serialized in a specific way to be passed as
an method argument or receive with an answer. An example is given here.
{{{
`%(filter)s`
}}}

=== Signal Objects ===

Signal is available for TCP connection only.
Signal object has been serialized in a specific way to be send to client.
An example is given here.
{{{
`%(signal)s`
}}}

== Common Available Commands ==

%(commands)s

== Http Specific Commands ==

=== General Commands ===

%(serverinfo_cmd)s

%(web_commands)s

== TCP Specific Commands ==

=== General Commands ===

%(close_cmd)s

%(tcp_commands)s
""" % {
        "cmd_format": self.commandDoc(),
        "answer": self.answerDoc(),
        "filter": Get_json_filter(filter).to_pretty_json(),
        "commands": "\n\n".join(map(self.formatSectionDoc, sections)),
        "signal": DeejaydJSONSignal(signal).to_pretty_json(),
        "web_commands": self.formatSectionDoc({\
                "prefix": "web.",
                "desc": "Commands specific to webui",
                "object": protocol.DeejaydWebJSONRPC,
            }),
        "tcp_commands": self.formatSectionDoc({\
                "prefix": "signal.",
                "desc": "Signal subscription commands",
                "object": protocol.DeejaydSignalJSONRPC,
            }),
        "serverinfo_cmd": self.formatCommandDoc(\
                protocol.DeejaydHttpJSONRPC.jsonrpc_serverInfo),
        "close_cmd": self.formatCommandDoc(\
                protocol.DeejaydTcpJSONRPC.jsonrpc_close),
    }

if __name__ == "__main__":
    docs = WikiFormat().build(common_request)

    f = open("doc/deejayd_rpc_protocol","w")
    try: f.write(docs)
    finally: f.close()


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os, threading, traceback, Queue
from deejayd.ui import log
from deejayd.utils import str_encode
import pyinotify

#############################################################################
##### Events Watcher
#############################################################################

def log_event(func):
    def log_event_func(self, event):
        path = os.path.join(event.path.decode("utf-8"),\
                event.name.decode("utf-8"))
        try:
            event.maskname
        except AttributeError:
            event.maskname = event.event_name
        log.info(_("Inotify event %s: %s") % (event.maskname, path))
        func(self,event)

    return log_event_func

class InotifyWatcher(pyinotify.ProcessEvent):

    def __init__(self, library, queue):
        self.__library = library
        self.__queue = queue

    @log_event
    def process_IN_CREATE(self, event):
        self.__queue.put(("create", self.__library, event))

    @log_event
    def process_IN_DELETE(self, event):
        self.__queue.put(("delete", self.__library, event))

    @log_event
    def process_IN_MOVED_FROM(self, event):
        self.__queue.put(("move_from", self.__library, event))

    @log_event
    def process_IN_MOVED_TO(self, event):
        self.__queue.put(("move_to", self.__library, event))

    @log_event
    def process_IN_CLOSE_WRITE(self, event):
        self.__queue.put(("close_write", self.__library, event))

    def process_IN_IGNORED(self, event):
        # This is said to be useless in the documentation, and is
        # effectively useless for us except for adding extreme verbosity
        # to the tests. Farewell, IN_IGNORED!
        pass


class _LibraryWatcher(threading.Thread):

    def __init__(self, db, queue):
        threading.Thread.__init__(self)
        self.should_stop = threading.Event()
        self.__db = db
        self.__queue = queue
        self.__created_files = []
        self.__need_update = False
        self.__record_changes = []

    def run(self):
        while not self.should_stop.isSet():
            try: type, library, event = self.__queue.get(True, 0.1)
            except Queue.Empty:
                continue
            try:
                changes = self.__execute(type,library,event)
                if changes:
                    self.__record_changes.extend(changes)
                    self.__need_update = True
            except Exception, ex:
                path = str_encode(os.path.join(event.path, event.name),
                                  errors='replace')
                log.err(_("Inotify problem for '%s', see traceback") % path)
                log.err("------------------Traceback lines--------------------")
                log.err(traceback.format_exc())
                log.err("-----------------------------------------------------")
            if self.__need_update and self.__queue.empty(): # record changes
                library.inotify_record_changes(self.__record_changes)
                self.__record_changes = []
                self.__need_update = False
        self.__db.close()

    def __occured_on_dirlink(self, library, event):
        if not event.name:
            return False
        file_path = os.path.join(event.path, event.name)
        if os.path.exists(file_path):
            return os.path.islink(file_path) and os.path.isdir(file_path)
        else:
            # File seems to have been deleted, so we lookup for a dirlink
            # in the library.
            return file_path in library.get_root_paths()

    def __execute(self, type, library, event):
        # first be sure that path are correct
        try:
            path = library._encode(event.path)
            name = library._encode(event.name)
        except UnicodeError: # skip this event
            return False

        if type == "create":
            if self.__occured_on_dirlink(library, event):
                return library.add_directory(path, name, True)
            elif not self.is_on_dir(event):
                self.__created_files.append((path, name))
        elif type == "delete":
            if self.__occured_on_dirlink(library, event):
                return library.remove_directory(path, name, True)
            elif not self.is_on_dir(event):
                return library.remove_file(path, name)
        elif type == "move_from":
            if not self.is_on_dir(event):
                return library.remove_file(path, name)
            else:
                return library.remove_directory(path, name)
        elif type == "move_to":
            if not self.is_on_dir(event):
                return library.add_file(path, name)
            else:
                return library.add_directory(path, name)
        elif type == "close_write":
            if (path, name) in self.__created_files:
                del self.__created_files[\
                        self.__created_files.index((path, name))]
                return library.add_file(path, name)
            else:
                return library.update_file(path, name)

        return False

    def close(self):
        self.should_stop.set()
        threading.Thread.join(self)


class LibraryWatcher(_LibraryWatcher):

    def is_on_dir(self, event):
        return event.dir


class LibraryWatcherOLD(_LibraryWatcher):

    def is_on_dir(self, event):
        return event.is_dir


#############################################################################

class _DeejaydInotify(threading.Thread):

    def __init__(self, db, audio_library, video_library):
        threading.Thread.__init__(self)
        self.should_stop = threading.Event()

        self.__audio_library = audio_library
        self.__video_library = video_library
        self.__db = db
        self.__queue = Queue.Queue(1000)

        self.__wm = pyinotify.WatchManager()
        self.__watched_dirs = {}
        self.EVENT_MASK = self.watched_events_mask()

    def is_watched (self, dir_path):
        return dir_path in self.__watched_dirs.keys()

    def watch_dir(self, dir_path, library):
        if self.is_watched(dir_path):
            raise ValueError('dir %s is already watched' % dir_path)
        wdd = self.__wm.add_watch(dir_path, self.EVENT_MASK,
                      proc_fun=InotifyWatcher(library,self.__queue), rec=True,
                      auto_add=True)
        self.__watched_dirs[dir_path] = wdd

    def stop_watching_dir(self, dir_path):
        if self.is_watched(dir_path):
            wdd = self.__watched_dirs[dir_path]
            del self.__watched_dirs[dir_path]
            self.__wm.rm_watch(wdd[dir_path], rec=True)

    def run(self):
        notifier = self.notifier(self.__wm)

        for library in (self.__audio_library, self.__video_library):
            if library:
                library.watcher = self
                for dir_path in library.get_root_paths():
                    self.watch_dir(dir_path, library)

        # start library watcher thread
        lib_watcher = self.watcher(self.__db, self.__queue)
        lib_watcher.start()
        while not self.should_stop.isSet():
            # process the queue of events as explained above
            notifier.process_events()
            if notifier.check_events():
                # read notified events and enqeue them
                notifier.read_events()
        lib_watcher.close()
        notifier.stop()

    def close(self):
        self.should_stop.set()
        threading.Thread.join(self)


class DeejaydInotify(_DeejaydInotify):

    def watched_events_mask(self):
        return pyinotify.IN_DELETE |\
               pyinotify.IN_CREATE |\
               pyinotify.IN_MOVED_FROM |\
               pyinotify.IN_MOVED_TO |\
               pyinotify.IN_CLOSE_WRITE

    def watcher(self, db, queue):
        return LibraryWatcher(db, queue)

    def notifier(self, watch_manager):
        return pyinotify.Notifier(watch_manager, timeout=1000)


class DeejaydInotifyOLD(_DeejaydInotify):

    def watched_events_mask(self):
        return pyinotify.EventsCodes.IN_DELETE |\
               pyinotify.EventsCodes.IN_CREATE |\
               pyinotify.EventsCodes.IN_MOVED_FROM |\
               pyinotify.EventsCodes.IN_MOVED_TO |\
               pyinotify.EventsCodes.IN_CLOSE_WRITE

    def watcher(self, db, queue):
        return LibraryWatcherOLD(db, queue)

    def notifier(self, watch_manager):
        return pyinotify.Notifier(watch_manager)


def get_watcher(db, audio_library, video_library):
    try:
        pyinotify_version = map(int, pyinotify.__version__.split('.'))
    except AttributeError:
        pyinotify_version = [0, 7]

    if pyinotify_version >= [0, 8]:
        return DeejaydInotify(db, audio_library, video_library)
    else:
        return DeejaydInotifyOLD(db, audio_library, video_library)


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.database import schema

def upgrade(cursor, backend, config):
    # create new table medialist_property
    for table in schema.db_schema:
        if table.name == "medialist_property":
            for stmt in backend.to_sql(table):
                cursor.execute(stmt)
            break

    # update db version
    sql = [
        "UPDATE variables SET value = '11' WHERE name = 'database_version';",
        ]
    for s in sql:
        cursor.execute(s)

# vim ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os,subprocess

from deejayd.component import SignalingComponent
from deejayd.plugins import PluginError, IPlayerPlugin
from deejayd.player import PlayerError
from deejayd.ui import log
from deejayd.utils import get_uris_from_pls, get_uris_from_m3u

PLAYER_PLAY = "play"
PLAYER_PAUSE = "pause"
PLAYER_STOP = "stop"

class UnknownPlayer(SignalingComponent):

    def __init__(self, db, plugin_manager, config):
        SignalingComponent.__init__(self)
        self.config = config
        self.db = db
        # Init plugins
        self.plugins = []
        plugins_cls = plugin_manager.get_plugins(IPlayerPlugin)
        for plugin in plugins_cls:
            try: self.plugins.append(plugin(config))
            except PluginError, err:
                log.err(_("Unable to init %s plugin: %s")%(plugin.NAME, err))

        # Initialise var
        self._video_support = False
        self._source = None
        self._media_file = None
        self._replaygain = config.getboolean("general","replaygain")


    def load_state(self):
        # Restore volume
        self.set_volume(float(self.db.get_state("volume")))
        # Restore current media
        media_pos = int(self.db.get_state("current"))
        source = self.db.get_state("current_source")
        if media_pos != -1 and source not in ("queue", "none", 'webradio'):
            self._media_file = self._source.get(media_pos, "pos", source)
        # Update state
        state = self.db.get_state("state")
        if state != PLAYER_STOP and source != 'webradio':
            try:
                self.play()
            except PlayerError:
                # There is an issue restoring the playing state on this file
                # so pause for now.
                self.stop()
            else:
                if self._media_file and self._media_file["source"] != "queue":
                    self.set_position(int(self.db.get_state("current_pos")))
        if state == PLAYER_PAUSE:
            self.pause()

    def init_video_support(self):
        self._video_support = True
        self._fullscreen = self.config.getboolean("general", "fullscreen")

    def set_source(self,source):
        self._source = source

    def play(self):
        if self.get_state() == PLAYER_STOP:
            file = self._media_file or self._source.get_current()
            self._change_file(file)
        elif self.get_state() in (PLAYER_PAUSE, PLAYER_PLAY):
            self.pause()

    def start_play(self):
        if not self._media_file: return
        if self._media_file["type"] == "webradio":
            if self._media_file["url-type"] == "pls":
                try:
                    urls = get_uris_from_pls(self._media_file["url"])
                except IOError:
                    raise PlayerError(_("Unable to get pls for webradio %s")\
                                      % self._media_file["title"])
                if not len(urls): # we don't succeed to extract uri
                    raise PlayerError(\
                            _("Unable to extract uri from pls playlist"))
                self._media_file.update({"urls": urls, "url-index": 0, \
                                         "url-type": "urls"})
            idx = self._media_file["url-index"]
            self._media_file["uri"] = \
                    self._media_file["urls"][idx].encode("utf-8")

    def pause(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def next(self):
        if self.get_state() != PLAYER_STOP:
            try: self._media_file.skip()
            except AttributeError: pass
        self._change_file(self._source.next())

    def previous(self):
        self._change_file(self._source.previous())

    def go_to(self,nb,type,source = None):
        self._change_file(self._source.get(nb,type,source))

    def get_volume(self):
        raise NotImplementedError

    def set_volume(self,v):
        raise NotImplementedError

    def get_position(self):
        raise NotImplementedError

    def set_position(self, pos, relative = False):
        if relative and self.get_state()!="stop"\
                and self._media_file["type"]!="webradio":
            cur_pos = self.get_position()
            pos = int(pos) + cur_pos
        self._set_position(pos)

    def _set_position(self, pos):
        raise NotImplementedError

    def get_state(self):
        raise NotImplementedError

    def set_option(self, name, value):
        if not self._media_file or self._media_file["type"] != "video" or\
                                   self.get_state() == PLAYER_STOP:
            return

        options = {
            "audio_lang": self.set_alang,
            "sub_lang": self.set_slang,
            "av_offset": self.set_avoffset,
            "sub_offset": self.set_suboffset,
            "zoom": self.set_zoom,
            "aspect_ratio": self.set_aspectratio,
            }
        options[name](value)

    def set_zoom(self, zoom):
        raise NotImplementedError

    def set_avoffset(self, offset):
        raise NotImplementedError

    def set_suboffset(self, offset):
        raise NotImplementedError

    def set_aspectratio(self, aspect):
        raise NotImplementedError

    def set_alang(self,lang_idx):
        try: audio_tracks = self._media_file["audio"]
        except KeyError:
            raise PlayerError(_("Current media hasn't multiple audio channel"))
        else:
            if lang_idx in (-2,-1): # disable/auto audio channel
                self._player_set_alang(lang_idx)
                self._media_file["audio_idx"] = self._player_get_alang()
                return
            found = False
            for track in audio_tracks:
                if track['ix'] == lang_idx: # audio track exists
                    self._player_set_alang(lang_idx)
                    found = True
                    break
            if not found:
                raise PlayerError(_("Audio channel %d not found") % lang_idx)
            self._media_file["audio_idx"] = self._player_get_alang()

    def set_slang(self,lang_idx):
        try: sub_tracks = self._media_file["subtitle"]
        except KeyError:
            raise PlayerError(_("Current media hasn't multiple sub channel"))
        else:
            if lang_idx in (-2,-1): # disable/auto subtitle channel
                self._player_set_slang(lang_idx)
                self._media_file["subtitle_idx"] = self._player_get_slang()
                return
            found = False
            for track in sub_tracks:
                if track['ix'] == lang_idx: # audio track exists
                    self._player_set_slang(lang_idx)
                    found = True
                    break
            if not found:
                raise PlayerError(_("Sub channel %d not found") % lang_idx)
            self._media_file["subtitle_idx"] = self._player_get_slang()

    def get_playing(self):
        return self.get_state() != PLAYER_STOP and self._media_file or None

    def is_playing(self):
        return self.get_state() != PLAYER_STOP

    def get_status(self):
        status = [("state",self.get_state()),("volume",self.get_volume())]

        if self._media_file:
            status.append(("current",\
              "%d:%s:%s" % (self._media_file["pos"],\
                    str(self._media_file["id"]), self._media_file["source"])))

        if self.get_state() != PLAYER_STOP:
            position = self.get_position()
            if "length" not in self._media_file.keys() or \
                                              self._media_file["length"] == 0:
                length = position
            else:
                length = int(self._media_file["length"])
            status.extend([ ("time","%d:%d" % (position, length)) ])

        return status

    def close(self):
        # close plugins
        for plugin in self.plugins:
            plugin.close()

        # save state
        current = self._media_file or {"pos": "-1", "source": "none"}
        states = [
            (str(self.get_volume()), "volume"),
            (current["pos"], "current"),
            (current["source"], "current_source"),
            (str(self.get_position()), "current_pos"),
            (self.get_state(), "state"),
            ]
        self.db.set_state(states)

        # stop player if necessary
        if self.get_state() != PLAYER_STOP:
            self.stop()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import time
from datetime import datetime
try: import json # python 2.6
except ImportError: # if python < 2.6, require simplejson
    import simplejson as json
from deejayd.rpc import *


class JSONRPCEncoder(json.JSONEncoder):
    """
    Provide custom serializers for JSON-RPC.
    """
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y%m%dT%H:%M:%S")
        raise TypeError("%r is not JSON serializable" % (obj,))


class _DeejaydJSON:

    def dump(self):
        return self._build_obj()

    def to_json(self):
        return json.dumps(self._build_obj(), cls=JSONRPCEncoder)

    def to_pretty_json(self):
        s = json.dumps(self._build_obj(), sort_keys=True, indent=4)
        return '\n'.join([l.rstrip() for l in  s.splitlines()])


class JSONRPCRequest(_DeejaydJSON):
    """
    Build JSON-RPC Request
    """
    def __init__(self, method_name, params, notification = False, id = None):
        self.method = method_name
        self.params = params
        # use timestamp as id if no id has been given
        self.id = None
        if not notification:
            self.id = id or int(time.time())

    def _build_obj(self):
        return {"method": self.method, "params": self.params, "id": self.id}

    def get_id(self):
        return self.id


class JSONRPCResponse(_DeejaydJSON):
    """
    Build JSON-RPC Response
    """
    def __init__(self, result, id):
        self.id = id
        self.result = result

    def _build_obj(self):
        result, error = self.result, None
        if isinstance(self.result, Fault):
            error = {"code": self.result.code, "message": str(self.result)}
            result = None

        return {"result": result, "error": error, "id": self.id}

    def to_json(self):
        try:
            return json.dumps(self._build_obj(), cls=JSONRPCEncoder)
        except TypeError, ex:
            error = {"code": NOT_WELLFORMED_ERROR, "message": str(ex)}
            obj = {"result": None, "error": error, "id": self.id}
            return json.dumps(obj)

#
# JSON filter serializer
#
class JSONFilter(_DeejaydJSON):

    def __init__(self, filter):
        self.filter = filter

    def _get_value(self):
        raise NotImplementedError

    def _build_obj(self):
        return {
            "type": self.type,
            "id": self.filter.get_identifier(),
            "value": self._get_value(),
            }

class JSONBasicFilter(JSONFilter):
    type = "basic"
    def _get_value(self):
        return {"tag": self.filter.tag, "pattern": self.filter.pattern}

class JSONComplexFilter(JSONFilter):
    type = "complex"
    def _get_value(self):
        return [Get_json_filter(f).dump() for f in self.filter.filterlist]

def Get_json_filter(filter):
    if filter is None:
        return None
    if filter.type == 'basic':
        json_filter_class = JSONBasicFilter
    elif filter.type == 'complex':
        json_filter_class = JSONComplexFilter
    return json_filter_class(filter)

#
# JSON signal serializer
#
class DeejaydJSONSignal(_DeejaydJSON):

    def __init__(self, signal):
        self.name = signal is not None and signal.get_name() or ""
        self.attrs = signal is not None and signal.get_attrs() or {}

    def set_name(self, name):
        self.name = name

    def _build_obj(self):
        return {"type": "signal",\
                "answer": {"name": self.name, "attrs": self.attrs}}

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

try: import json # python 2.6
except ImportError: # if python < 2.6, require simplejson
    import simplejson as json
from deejayd.mediafilters import *
from deejayd.rpc import *


def loads_request(string, **kws):
    err = Fault(NOT_WELLFORMED_ERROR, "Bad json-rpc request")

    try: unmarshalled = json.loads(string, **kws)
    except ValueError:
        raise err

    if (isinstance(unmarshalled, dict)):
        for key in ("method", "params", "id"):
            if key not in unmarshalled:
                raise err
        return unmarshalled
    raise err

def loads_response(string, **kws):
    err = Fault(NOT_WELLFORMED_ERROR, "Bad json-rpc response")

    try: ans = json.loads(string, **kws)
    except ValueError:
        raise err

    for key in ("error", "result", "id"):
        if key not in ans:
            raise err
    return ans


def Parse_json_filter(json_filter):
    try:
        name = json_filter["id"]
        type = json_filter["type"]
        if type == "basic":
            filter_class = NAME2BASIC[name]
            filter = filter_class(json_filter["value"]["tag"], \
                    json_filter["value"]["pattern"])
        elif type == "complex":
            filter = NAME2COMPLEX[name]()
            for f in json_filter["value"]:
                filter.combine(Parse_json_filter(f))
        else:
            raise TypeError
    except (KeyError, TypeError):
        raise Fault(NOT_WELLFORMED_ERROR,\
                "Bad filter argument for this json-rpc request")

    return filter

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from deejayd.utils import quote_uri
import kaa.metadata


class _MediaFile(object):
    type = "unknown"

    def parse(self, file_path):
        return {
            "filename": os.path.basename(file_path),
            "uri": quote_uri(file_path),
            "type": self.type,
            "rating": "2", # [0-4]
            "lastplayed": "0",
            "skipcount": "0",
            "playcount": "0",
            }


class _AudioFile(_MediaFile):
    _tagclass_ = None
    type = "song"
    supported_tag = ("tracknumber","title","genre","artist","album",\
            "discnumber","date","replaygain_track_gain",\
            "replaygain_track_peak")

    def _format_number(self, nb):
        numbers = nb.split("/")
        try: numbers = ["%02d" % int(num) for num in numbers]
        except (TypeError, ValueError):
            return nb

        return "/".join(numbers)

    def parse(self, file):
        infos = _MediaFile.parse(self, file)
        if self._tagclass_:
            tag_info = self._tagclass_(file)
            for i in ("bitrate", "length"):
                try: infos[i] = int(getattr(tag_info.info, i))
                except AttributeError:
                    infos[i] = 0
            for t in self.supported_tag:
                try: info = tag_info[t][0]
                except:
                    info = ''
                if t in ("tracknumber", "discnumber"):
                    info = self._format_number(info)
                infos[t] = info
            # get front cover album if available
            infos["various_artist"] = infos["artist"]
            cover = self.get_cover(tag_info)
            if cover: infos["cover"] = cover

        return infos

class _VideoFile(_MediaFile):
    type = "video"
    mime_type = None

    def _format_title(self, f):
        (filename,ext) = os.path.splitext(f)
        title = filename.replace(".", " ")
        title = title.replace("_", " ")
        return title.title()

    def parse(self, file):
        infos = _MediaFile.parse(self, file)
        infos.update({
                "audio_channels": "0",
                "subtitle_channels": "0",
                "length": "0",
                "videoheight": "0",
                "videowidth": "0",
                })
        (path,filename) = os.path.split(file)
        infos["title"] = self._format_title(filename)

        # parse video file with kaa
        kaa_infos = kaa.metadata.parse(file)
        if kaa_infos is None:
            raise TypeError(_("Video media %s not supported by kaa parser") \
                    % file)
        if len(kaa_infos["video"]) == 0:
            raise TypeError(_("This file is not a video"))
        infos["length"] = int(kaa_infos["length"])
        infos["videowidth"] = kaa_infos["video"][0]["width"]
        infos["videoheight"] = kaa_infos["video"][0]["height"]
        infos["audio_channels"] = len(kaa_infos["audio"])
        infos["subtitle_channels"] = len(kaa_infos["subtitles"])

        return infos

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.database import schema

def upgrade(cursor, backend, config):
    # get covers
    cursor.execute("SELECT * FROM cover;")
    covers = cursor.fetchall()

    # drop old table
    cursor.execute("DROP TABLE cover")
    # create new table
    for table in schema.db_schema:
        if table.name != "cover": continue
        for stmt in backend.to_sql(table):
            cursor.execute(stmt)

    query = "INSERT INTO cover (id,source,mime_type,lmod,image)\
            VALUES(%s,%s,%s,%s,%s)"
    for (id, source, lmod, img) in covers:
        mime = "image/jpeg"
        cursor.execute(query, (id, source, mime, lmod, img))

    sql = [
        "UPDATE variables SET value = '9' WHERE name = 'database_version';",
        ]
    for s in sql:
        cursor.execute(s)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
import ctypes


try:
    _libXext = ctypes.cdll.LoadLibrary('libXext.so.6')
except (ImportError, OSError), e:
    raise ImportError, e


# Bool DPMSQueryExtension (Display *display, int *event_base, int *error_base)
_libXext.DPMSQueryExtension.argtypes = (ctypes.c_void_p,
                                        ctypes.POINTER(ctypes.c_int),
                                        ctypes.POINTER(ctypes.c_int))
_libXext.DPMSQueryExtension.restype = ctypes.c_int

# Status DPMSEnable (Display *display )
_libXext.DPMSEnable.argtypes = (ctypes.c_void_p, )
_libXext.DPMSEnable.restype = ctypes.c_int

# Status DPMSDisable (Display *display )
_libXext.DPMSDisable.argtypes = (ctypes.c_void_p, )
_libXext.DPMSDisable.restype = ctypes.c_int


# copy functions from the library
module = sys.modules[__name__]
for name in dir(_libXext):
    if name.startswith('DPMS'):
        setattr(module, name, getattr(_libXext, name))


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.interfaces import DeejaydError
from deejayd.ui import log


AVAILABLE_BACKENDS = ('xine', 'gstreamer', )


class PlayerError(DeejaydError):pass

def init(db, plugin_manager, config):
    media_backend = config.get("general","media_backend")

    if media_backend == "auto":
        backend_it = iter(AVAILABLE_BACKENDS)
        media_backend = None
        try:
            while not media_backend:
                backend = backend_it.next()
                try:
                    __import__('.'.join(('deejayd', 'player', backend, )))
                except ImportError:
                    # Do nothing, simply ignore
                    pass
                else:
                    media_backend = backend
                    config.set('general', 'mediabackend', backend)
                    log.msg(_("Autodetected %s backend." % backend))
        except StopIteration:
            log.err(_("Could not find suitable media backend."), fatal=True)

    if media_backend == "gstreamer":
        from deejayd.player import gstreamer
        try: player = gstreamer.GstreamerPlayer(db, plugin_manager, config)
        except PlayerError, err:
            log.err(str(err), fatal=True)

    elif media_backend == "xine":
        from deejayd.player import xine
        try: player = xine.XinePlayer(db, plugin_manager, config)
        except PlayerError, err:
            log.err(str(err), fatal=True)

    else: log.err(_("Invalid media backend"), fatal=True)

    return player

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""A generic resource for publishing objects via JSON-RPC.
Requires simplejson; can be downloaded from
http://cheeseshop.python.org/pypi/simplejson
"""
from __future__ import nested_scopes

# System Imports
import urlparse

# Sibling Imports
from twisted.web import resource, server
from twisted.internet import defer, protocol, reactor
from twisted.web import http

from deejayd.ui import log
from deejayd.rpc import Fault
from deejayd.rpc.jsonparsers import loads_request
from deejayd.rpc.jsonbuilders import JSONRPCResponse
from deejayd.rpc import protocol as deejayd_protocol


class Handler:
    """Handle a JSON-RPC request and store the state for a request in progress.

    Override the run() method and return result using self.result,
    a Deferred.

    We require this class since we're not using threads, so we can't
    encapsulate state in a running function if we're going  to have
    to wait for results.

    For example, lets say we want to authenticate against twisted.cred,
    run a LDAP query and then pass its result to a database query, all
    as a result of a single JSON-RPC command. We'd use a Handler instance
    to store the state of the running command.
    """

    def __init__(self, resource, *args):
        self.resource = resource # the JSON-RPC resource we are connected to
        self.result = defer.Deferred()
        self.run(*args)

    def run(self, *args):
        # event driven equivalent of 'raise UnimplementedError'
        self.result.errback(NotImplementedError("Implement run() in subclasses"))


class JSONRPC(resource.Resource, deejayd_protocol.DeejaydHttpJSONRPC):
    """A resource that implements JSON-RPC.

    Methods published can return JSON-RPC serializable results, Faults,
    Binary, Boolean, DateTime, Deferreds, or Handler instances.

    By default methods beginning with 'jsonrpc_' are published.

    Sub-handlers for prefixed methods (e.g., system.listMethods)
    can be added with putSubHandler. By default, prefixes are
    separated with a '.'. Override self.separator to change this.
    """

    # Error codes for Twisted, if they conflict with yours then
    # modify them at runtime.
    NOT_FOUND = 8001
    FAILURE = 8002

    isLeaf = 1

    def __init__(self, deejayd):
        super(JSONRPC, self).__init__()
        self.subHandlers = {}
        self.deejayd_core = deejayd

    def render(self, request):
        request.content.seek(0, 0)
        # Unmarshal the JSON-RPC data
        content = request.content.read()
        try:
            parsed = loads_request(content)
            args, functionPath = parsed['params'], parsed["method"]
            function = self._getFunction(functionPath)
        except Fault, f:
            try: id = parsed["id"]
            except:
                id = None
            self._cbRender(f, request, id)
        else:
            request.setHeader("content-type", "text/json")
            defer.maybeDeferred(function, *args).addErrback(
                self._ebRender, parsed["id"]
            ).addCallback(
                self._cbRender, request, parsed["id"]
            )
        return server.NOT_DONE_YET

    def _cbRender(self, result, request, id):
        if isinstance(result, Handler):
            result = result.result
        # build json answer
        ans = JSONRPCResponse(result, id).to_json()
        request.setHeader("content-length", str(len(ans)))
        request.write(ans)
        request.finish()

    def _ebRender(self, failure, id):
        if isinstance(failure.value, Fault):
            return failure.value
        log.err(failure)
        return Fault(self.FAILURE, "error")

__all__ = ["JSONRPC", "Handler"]

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from os import path
from ConfigParser import NoOptionError
from deejayd.ui import log
from deejayd.database.queries import DatabaseQueries
from deejayd.database import schema

DatabaseError = None

def init(config):
    global DatabaseError

    db_type =  config.get("database","db_type")
    db_name = config.get("database","db_name")
    try: backend = __import__('deejayd.database.backends.%s' % db_type,\
                              {}, {}, [''])
    except ImportError, ex:
        log.err(_(\
      "You chose a database which is not supported. see config file. %s")\
      % str(ex), fatal = True)
    DatabaseError = backend.DatabaseError

    if db_type == "sqlite":
        connection = backend.DatabaseWrapper(db_name)
    elif db_type == "mysql":
        db_user = config.get("database","db_user")
        db_password = config.get("database","db_password")
        try: db_host = config.get("database","db_host")
        except NoOptionError:
            db_host = ""
        try: db_port = config.getint("database","db_port")
        except (NoOptionError, ValueError):
            db_port = 3306

        connection = backend.DatabaseWrapper(db_name, db_user, db_password,\
            db_host, db_port)

    # verify database version
    cursor = connection.cursor()
    try:
        cursor.execute("SELECT value FROM variables\
                            WHERE name = 'database_version'")
        (db_version,) = cursor.fetchone()
        db_version = int(db_version)
    except DatabaseError: # initailise db
        for table in schema.db_schema:
            for stmt in backend.to_sql(table):
                cursor.execute(stmt)
        for query in backend.custom_queries:
            cursor.execute(query)
        log.info(_("Database structure successfully created."))
        for query in schema.db_init_cmds:
            cursor.execute(query)
        log.info(_("Initial entries correctly inserted."))

        DatabaseQueries.structure_created = True
        connection.commit()
    else:
        if schema.db_schema_version > db_version:
            log.info(_("The database structure needs to be updated..."))

            base = path.dirname(__file__)
            base_import = "deejayd.database.upgrade"
            i = db_version+1
            while i < schema.db_schema_version+1:
                db_file = "db_%d" % i
                try: up = __import__(base_import+"."+db_file, {}, {}, base)
                except ImportError:
                    err = _("Unable to upgrade database, have to quit")
                    log.err(err, True)
                up.upgrade(cursor, backend, config)
                i += 1
            connection.commit()

    cursor.close()
    return DatabaseQueries(connection)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os, glob, ConfigParser
from zope.interface import Interface, Attribute
from deejayd.interfaces import DeejaydError

class PluginError(DeejaydError): pass

class PluginManager(object):

    def __init__(self, config):
        try:
            self.enabled_plugins = config.getlist("general", "enabled_plugins")
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            self.enabled_plugins = []

    def get_plugins(self, interface):
        plugins = []

        base = os.path.dirname(__file__)
        base_import = "deejayd.plugins"
        modules = [os.path.basename(f[:-3]) \
                   for f in glob.glob(os.path.join(base, "[!_]*.py"))]
        for m in modules:
            mod = __import__(base_import+"."+m, {}, {}, base)
            for cls_name in dir(mod):
                cls = getattr(mod, cls_name)
                try:
                    if interface.implementedBy(cls) \
                        and cls.NAME in self.enabled_plugins:
                        plugins.append(cls)
                except:
                    continue

        return plugins


class IWebradioPlugin(Interface):
    NAME = Attribute("Name of the plugin")
    HAS_CATEGORIE = Attribute("Set to true if this module support categorie")

    def get_categories(self):
        """ return list of categories supported by this plugin
            raise an exception if no categorie has been supported"""

    def get_streams(self, categorie = None):
        """ return list of streams for a given categories """

class IPlayerPlugin(Interface):
    NAME = Attribute("Name of the plugin")

    def on_media_played(self, media):
        """ Call when a track has been played """

    def close(self):
        """ Call when we closed the player """

# vim: ts=4 sw=4 expandtab


# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os,re

from deejayd.xmlobject import DeejaydXMLObject, ET
from deejayd.utils import *


class RdfBuilder(DeejaydXMLObject):

    def __init__(self, source_name):
        self.rdf_nsp = "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}"
        self.file_nsp = "{http://%s/rdf#}" % source_name
        self.xmlroot = ET.Element(self.rdf_nsp+"RDF")

    def build_seq(self, url, parent = None):
        if parent is None: parent = self.xmlroot
        seq = ET.SubElement(parent, self.rdf_nsp+"Seq")
        seq.attrib[self.rdf_nsp+"about"] = self._to_xml_string(url)

        return seq

    def build_li(self, parent, url = None):
        li = ET.SubElement(parent, self.rdf_nsp+"li")
        if url: li.attrib[self.rdf_nsp+"about"] = self._to_xml_string(url)

        return li

    def build_item_desc(self, parms, parent = None, url = None):
        if parent is None: parent = self.xmlroot
        desc = ET.SubElement(parent, self.rdf_nsp+"Description")
        if url: desc.attrib[self.rdf_nsp+"about"] = self._to_xml_string(url)

        for p in parms.keys():
            if p in ("time","length"):
                if parms[p]:
                    value = format_time(int(parms[p]))
                else:
                    value = self._to_xml_string(0)
            elif p == "external_subtitle":
                value = parms[p] == "" and _("No") or _("Yes")
            elif p == "rating":
                rating = u'\u266a' * int(parms[p])
                value = self._to_xml_string(rating)
            else:
                try: value = self._to_xml_string(parms[p])
                except TypeError:
                    continue
            node = ET.SubElement(desc, self.file_nsp+"%s"\
                % self._to_xml_string(p))
            node.text = value

        return desc

    def set_resource(self, elt, url):
        elt.attrib[self.rdf_nsp+"resource"] = self._to_xml_string(url)

    def to_xml(self):
        return '<?xml version="1.0" encoding="utf-8"?>' + "\n" + \
                ET.tostring(self.xmlroot)


class _DeejaydSourceRdf(DeejaydXMLObject):
    name = "unknown"
    locale_strings = ("%d Song", "%d Songs")

    def __init__(self, deejayd, rdf_dir):
        self._deejayd = deejayd
        self._rdf_dir = rdf_dir

    def update(self):
        current_id = self._get_current_id()
        try:
            status = self._deejayd.get_status(objanswer=False)
            new_id = status[self.__class__.name]
        except KeyError:
            return {"desc": ""}# this mode is not active
        new_id = int(new_id) % 10000
        # get media list
        obj = getattr(self._deejayd, self.__class__.get_list_func)()
        res = obj.get(objanswer=False)
        if isinstance(res, tuple):
            medias, filter, sort = res
        else:
            medias, filter, sort = res, None, None

        if current_id != new_id:
            self._build_rdf_file(medias, new_id)
        # build description
        single, plural = self.__class__.locale_strings
        len = status[self.__class__.name + "length"]
        desc = ngettext(single,plural,int(len))%int(len)
        try: time = int(status[self.__class__.name + "timelength"])
        except KeyError:
            pass
        else:
            if time > 0: desc += " (%s)" % format_time_long(time)
        return {"desc": desc, "sort": sort}

    def _build_rdf_file(self, media_list, new_id):
        # build xml
        rdf_builder = RdfBuilder(self.__class__.name)
        seq = rdf_builder.build_seq("http://%s/all-content" % self.name)
        for media in media_list:
            li = rdf_builder.build_li(seq)
            rdf_builder.build_item_desc(media, li,\
                "http://%s/%s" % (self.name, media["id"]))

        self._save_rdf(rdf_builder, new_id)

    def _save_rdf(self, rdf_builder, new_id, name = None):
        name = name or self.__class__.name
        # first clean rdf dir
        for file in os.listdir(self._rdf_dir):
            path = os.path.join(self._rdf_dir,file)
            if os.path.isfile(path) and file.startswith(name+"-"):
                os.unlink(path)

        filename = "%s-%d.rdf" % (name, new_id);
        file_path = os.path.join(self._rdf_dir,filename)

        fd = open(file_path, "w")
        fd.write(rdf_builder.to_xml())
        fd.close()
        os.chmod(file_path,0644)

    def _get_current_id(self):
        ids = []
        for file in os.listdir(self._rdf_dir):
            if re.compile("^"+self.name+"-[0-9]+\.rdf$").search(file):
                t = file.split("-")[1] # id.rdf
                t = t.split(".")
                try : ids.append(int(t[0]))
                except ValueError: pass

        if ids == []: return 0
        else: return max(ids)

class DeejaydPlaylistRdf(_DeejaydSourceRdf):
    name = "playlist"
    get_list_func = "get_playlist"

class DeejaydPanelRdf(_DeejaydSourceRdf):
    name = "panel"
    get_list_func = "get_panel"

class DeejaydQueueRdf(_DeejaydSourceRdf):
    name = "queue"
    get_list_func = "get_queue"

class DeejaydWebradioRdf(_DeejaydSourceRdf):
    name = "webradio"
    locale_strings = ("%d Webradio", "%d Webradios")
    get_list_func = "get_webradios"

    def _build_rdf_file(self, wb_list, new_id):
        rdf_builder = RdfBuilder(self.__class__.name)
        seq = rdf_builder.build_seq("http://webradio/all-content")
        for wb in wb_list:
            if wb["url-type"] == "urls":
                urls = " | ".join(wb["urls"])
            else:
                urls = wb["url"]
            wb_item = rdf_builder.build_li(seq)
            rdf_builder.build_item_desc({"title": wb["title"], "url": urls},\
                    wb_item, url = "http://webradio/%s" % str(wb["id"]))

        self._save_rdf(rdf_builder, new_id)

class DeejaydVideoRdf(_DeejaydSourceRdf):
    name = "video"
    locale_strings = ("%d Video", "%d Videos")
    get_list_func = "get_video"

class DeejaydVideoDirRdf(_DeejaydSourceRdf):
    name = "videodir"

    def update(self):
        try:
            stats = self._deejayd.get_stats()
            new_id = stats["video_library_update"]
        except KeyError:
            return {"id": None}# this mode is not active
        current_id = self._get_current_id()
        new_id = int(new_id) % 10000
        if current_id != new_id:
            self._build_rdf_file(new_id)
        return {"id": new_id}

    def _build_rdf_file(self,new_id):
        rdf_builder = RdfBuilder(self.__class__.name)

        seq = rdf_builder.build_seq("http://videodir/all-content")
        self._build_dir_list(rdf_builder, seq, "")
        self._save_rdf(rdf_builder, new_id)

    def _build_dir_list(self, rdf_builder, seq_elt, dir, id = "1"):
        dir_elt = rdf_builder.build_li(seq_elt)
        dir_url = "http://videodir/%s" % os.path.join("root", dir)
        title = dir == "" and _("Root Directory") or os.path.basename(dir)
        rdf_builder.build_item_desc({"title": title}, url = dir_url)

        subdirs = self._deejayd.get_video_dir(dir).get_directories()
        if subdirs == []:
            rdf_builder.set_resource(dir_elt, dir_url)
        else:
            subdir_list =  rdf_builder.build_seq(dir_url, parent = dir_elt)
            for idx, d in enumerate(subdirs):
                new_id = id + "/%d" % (idx+1,)
                self._build_dir_list(rdf_builder, subdir_list,\
                    os.path.join(dir, d), new_id)

class DeejaydDvdRdf(_DeejaydSourceRdf):
    name = "dvd"
    locale_strings = ("%d Track", "%d Tracks")

    def update(self):
        current_id = self._get_current_id()
        try:
            status = self._deejayd.get_status(objanswer=False)
            new_id = status[self.__class__.name]
        except KeyError:
            return {"desc": ""}# this mode is not active
        dvd_content = self._deejayd.get_dvd_content(objanswer=False)
        new_id = int(new_id) % 10000
        if current_id != new_id:
            self._build_rdf_file(dvd_content, new_id)
        # build description
        single, plural = self.__class__.locale_strings
        len = status[self.__class__.name + "length"]
        desc = ngettext(single,plural,int(len))%int(len)
        return {"desc": desc, "title": dvd_content["title"],\
                "longest_track": dvd_content["longest_track"]}

    def _build_rdf_file(self, dvd_content, new_id):
        rdf_builder = RdfBuilder(self.__class__.name)

        # dvd structure
        seq = rdf_builder.build_seq("http://dvd/all-content")
        for track in dvd_content["track"]:
            track_li = rdf_builder.build_li(seq)

            track_url = "http://dvd/%s" % track["ix"]
            track_struct = rdf_builder.build_seq(track_url,track_li)

            track_data = {"title": _("Title %s") % track["ix"],\
                             "id": track["ix"],\
                         "length": track["length"]}
            rdf_builder.build_item_desc(track_data,None,track_url)

            for chapter in track["chapter"]:
                chapter_url = track_url + "/%s" % chapter["ix"]
                chapter_li = rdf_builder.build_li(track_struct)
                rdf_builder.set_resource(chapter_li,chapter_url)

                chapter_data = {"title": _("Chapter %s") % chapter["ix"],\
                                   "id": chapter["ix"],\
                               "length": chapter["length"]}
                rdf_builder.build_item_desc(chapter_data,None,chapter_url)

        self._save_rdf(rdf_builder,new_id)


ngettext("%d Song", "%d Songs", 0)
ngettext("%d Video", "%d Videos", 0)
ngettext("%d Webradio", "%d Webradios", 0)
ngettext("%d Track", "%d Tracks", 0)

modes = {
        "playlist": DeejaydPlaylistRdf,
        "queue": DeejaydQueueRdf,
        "panel": DeejaydPanelRdf,
        "webradio": DeejaydWebradioRdf,
        "video": DeejaydVideoRdf,
        "dvd": DeejaydDvdRdf,
        "videodir": DeejaydVideoDirRdf,
    }

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os

from deejayd.component import SignalingComponent
from deejayd.ui import log
from deejayd.player import PlayerError
from deejayd.sources._base import SourceError

class UnknownSourceException: pass

def format_rsp(func):
    def format_rsp_func(*__args, **__kw):
        (media, source) = func(*__args, **__kw)
        if media is not None:
            media["source"] = source
        return media

    return format_rsp_func


class SourceFactory(SignalingComponent):
    sources_list = ("playlist","queue","webradio","video","dvd","panel")

    def __init__(self,player,db,audio_library,video_library,plg_mnt,config):
        SignalingComponent.__init__(self)
        self.current = ""
        self.db = db
        activated_sources = config.getlist('general', "activated_modes")

        self.sources_obj = {}
        from deejayd.sources import queue
        self.sources_obj["queue"] = queue.QueueSource(db, audio_library)
        # playlist
        if "playlist" in activated_sources:
            from deejayd.sources import playlist
            self.sources_obj["playlist"] = playlist.PlaylistSource(db,\
                                                                 audio_library)
        else:
            log.info(_("Playlist support disabled"))

        # panel
        if "panel" in activated_sources:
            from deejayd.sources import panel
            self.sources_obj["panel"] = panel.PanelSource(db, audio_library,\
                    config)
        else:
            log.info(_("Panel support disabled"))

        # Webradio
        if "webradio" in activated_sources and player.is_supported_uri("http"):
            from deejayd.sources import webradio
            self.sources_obj["webradio"] = webradio.WebradioSource(db, plg_mnt)
        else:
            log.info(_("Webradio support disabled"))

        # init video support in player
        if "video" in activated_sources or "dvd" in activated_sources:
            try: player.init_video_support()
            except PlayerError:
                # Critical error, we have to quit deejayd
                msg = _('Cannot initialise video support, either disable video and dvd mode or check your player video support.')
                log.err(msg, fatal = True)
            except NotImplementedError:
                # player not supported video playback, quit deejayd
                msg = _("player '%s' don't support video playback, either disable video and dvd mode or change your player to have video support.")
                log.err(msg % player.name, fatal = True)

        # Video
        if "video" in activated_sources:
            from deejayd.sources import video
            self.sources_obj["video"] = video.VideoSource(db, video_library)
        else:
            log.info(_("Video support disabled"))

        # dvd
        if "dvd" in activated_sources and player.is_supported_uri("dvd"):
            from deejayd.sources import dvd
            try: self.sources_obj["dvd"] = dvd.DvdSource(db,config)
            except dvd.DvdError, ex:
                log.err(_("Unable to init dvd support : %s") % str(ex))
        else:
            log.info(_("DVD support disabled"))

        # restore recorded source
        source = db.get_state("source")
        try: self.set_source(source)
        except UnknownSourceException:
            log.err(_("Unable to set recorded source %s") % str(source))
            self.set_source(self.get_available_sources()[0])

        player.set_source(self)
        player.load_state()

    def set_option(self, source, name, value):
        try: self.sources_obj[source].set_option(name, value)
        except KeyError:
            raise UnknownSourceException
        except NotImplementedError:
            raise SourceError(_("option %s not supported for this mode"))
        self.dispatch_signame('player.status')

    def get_source(self,s):
        if s not in self.sources_obj.keys():
            raise UnknownSourceException

        return self.sources_obj[s]

    def set_source(self,s):
        if s not in self.sources_obj.keys():
            raise UnknownSourceException

        self.current = s
        self.dispatch_signame('mode')
        return True

    def get_status(self):
        status = [("mode",self.current)]
        for k in self.sources_obj.keys():
            status.extend(self.sources_obj[k].get_status())
        return status

    def get_all_sources(self):
        return [m for m in self.sources_list if m != "queue"]

    def get_available_sources(self):
        modes = self.sources_obj.keys()
        modes.remove("queue")
        return modes

    def is_available(self, mode):
        return mode in self.sources_obj.keys()

    def close(self):
        self.db.set_state([(self.current,"source")])
        for k in self.sources_obj.keys():
            self.sources_obj[k].close()

    #
    # Functions called from the player
    #
    @format_rsp
    def get(self, nb = None, type = "id", source_name = None):
        src = source_name or self.current
        return (self.sources_obj[src].go_to(nb,type), src)

    @format_rsp
    def get_current(self):
        queue_media = self.sources_obj["queue"].get_current() or \
                      self.sources_obj["queue"].next()
        if queue_media: return (queue_media, "queue")

        current = self.sources_obj[self.current].get_current() or \
            self.sources_obj[self.current].next(explicit = False)
        return (current, self.current)

    @format_rsp
    def next(self, explicit = True):
        queue_media = self.sources_obj["queue"].next(explicit)
        if queue_media: return (queue_media, "queue")
        return (self.sources_obj[self.current].next(explicit),self.current)

    @format_rsp
    def previous(self):
        return (self.sources_obj[self.current].previous(),self.current)

    def queue_reset(self):
        self.sources_obj["queue"].reset()


def init(player,db,audio_library,video_library,pl_mngt,config):
    source = SourceFactory(player,db,audio_library,video_library,pl_mngt,config)
    return source


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import __builtin__
import gettext

class DeejaydTranslations(gettext.GNUTranslations):

    def __init__(self, *args, **kwargs):
        self._catalog = {}
        self.plural = lambda n: n > 1
        gettext.GNUTranslations.__init__(self, *args, **kwargs)

    def install(self, unicode = True):
        if unicode:
            __builtin__.__dict__["_"] = self.ugettext
            __builtin__.__dict__["ngettext"] = self.ungettext
        else:
            __builtin__.__dict__["_"] = self.gettext
            __builtin__.__dict__["ngettext"] = self.ngettext

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import time
import pygst
pygst.require('0.10')

import gobject
import gst
import gst.interfaces

from ConfigParser import NoOptionError
from deejayd.player import PlayerError
from deejayd.player._base import *
from deejayd.ui import log

class GstreamerPlayer(UnknownPlayer):
    name = 'gstreamer'

    def __init__(self, db, plugin_manager, config):
        UnknownPlayer.__init__(self, db, plugin_manager, config)
        self.__volume = 100

        # Open a Audio pipeline
        pipeline =  self.config.get("gstreamer", "audio_output")
        if pipeline in ("gconf", "auto"):
            pipeline += "audiosink"
        else:
            pipeline += "sink"
        try: audio_sink = gst.parse_launch(pipeline)
        except gobject.GError, err:
            raise PlayerError(_("No audio sink found for Gstreamer : %s"%err))

        # More audio-sink option
        if pipeline == "alsasink":
            try: alsa_card = self.config.get("gstreamer", "alsa_card")
            except NoOptionError: pass
            else: audio_sink.set_property('device',alsa_card)

        self.bin = gst.element_factory_make('playbin')
        self.bin.set_property('video-sink', None)
        self.bin.set_property('audio-sink',audio_sink)

        self.bin.set_property("auto-flush-bus",True)
        bus = self.bin.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_message)

    def init_video_support(self):
        raise NotImplementedError

    def on_message(self, bus, message):
        if message.type == gst.MESSAGE_EOS:
            if self._media_file["type"] == "webradio":
                # an error happened, try the next url
                if self._media_file["url-index"] \
                        < len(self._media_file["urls"])-1:
                    self._media_file["url-index"] += 1
                    self._media_file["uri"] = \
                            self._media_file["urls"]\
                                [self._media_file["url-index"]].encode("utf-8")
                    self.start_play()
                return False
            elif self._media_file:
                try: self._media_file.played()
                except AttributeError: pass
                for plugin in self.plugins:
                    plugin.on_media_played(self._media_file)
            self._change_file(self._source.next(explicit = False))
        elif message.type == gst.MESSAGE_TAG:
            self._update_metadata(message.parse_tag())
        elif message.type == gst.MESSAGE_ERROR:
            err, debug = message.parse_error()
            err = str(err).decode("utf8", 'replace')
            log.err("Gstreamer : " + err)

        return True

    def start_play(self):
        super(GstreamerPlayer, self).start_play()
        if not self._media_file: return

        def open_uri(uri):
            self.bin.set_property('uri',uri)
            state_ret = self.bin.set_state(gst.STATE_PLAYING)
            timeout = 4
            state = None

            while state_ret == gst.STATE_CHANGE_ASYNC and timeout > 0:
                state_ret,state,pending_state = self.bin.get_state(1 * gst.SECOND)
                timeout -= 1

            if state_ret != gst.STATE_CHANGE_SUCCESS:
                msg = _("Unable to play file %s") % uri
                raise PlayerError(msg)

        if self._media_file["type"] == "webradio":
            while True:
                try: open_uri(self._media_file["uri"])
                except PlayerError, ex:
                    if self._media_file["url-index"] < \
                                            len(self._media_file["urls"])-1:
                        self._media_file["url-index"] += 1
                        self._media_file["uri"] = \
                                self._media_file["urls"]\
                                [self._media_file["url-index"]].encode("utf-8")
                    else:
                        raise ex
                else:
                    break
        else:
            try: open_uri(self._media_file["uri"])
            except PlayerError, ex:
                self._destroy_stream()
                raise ex

    def pause(self):
        if self.get_state() == PLAYER_PLAY:
            self.bin.set_state(gst.STATE_PAUSED)
        elif self.get_state() == PLAYER_PAUSE:
            self.bin.set_state(gst.STATE_PLAYING)
        else: return
        self.dispatch_signame('player.status')

    def stop(self,widget = None, event = None):
        self._media_file = None
        self.bin.set_state(gst.STATE_NULL)
        # FIXME : try to remove this one day ...
        self._source.queue_reset()

        self.dispatch_signame('player.status')

    def _change_file(self,new_file):
        sig = self.get_state() == PLAYER_STOP and True or False
        self.stop()
        self._media_file = new_file
        self.start_play()

        # replaygain reset
        self.set_volume(self.__volume, sig=False)

        if sig: self.dispatch_signame('player.status')
        self.dispatch_signame('player.current')

    def get_volume(self):
        return self.__volume

    def set_volume(self, vol, sig=True):
        self.__volume = min(100, int(vol))
        v = float(self.__volume)/100
        # replaygain support
        if self._replaygain and self._media_file is not None:
            try: scale = self._media_file.replay_gain()
            except AttributeError: pass # replaygain not supported
            else:
                v = max(0.0, min(4.0, v * scale))
                v = float(min(100, int(v * 100)))/100
        self.bin.set_property('volume', v)
        if sig: self.dispatch_signame('player.status')

    def get_position(self):
        if gst.STATE_NULL != self.__get_gst_state() and \
                self.bin.get_property('uri'):
            try: p = self.bin.query_position(gst.FORMAT_TIME)[0]
            except gst.QueryError: return 0
            p //= gst.SECOND
            return p
        return 0

    def _set_position(self,pos):
        if gst.STATE_NULL != self.__get_gst_state() and \
                self.bin.get_property('uri'):
            pos = max(0, int(pos))
            gst_time = pos * gst.SECOND

            event = gst.event_new_seek(
                1.0, gst.FORMAT_TIME, gst.SEEK_FLAG_FLUSH | \
                gst.SEEK_FLAG_ACCURATE, gst.SEEK_TYPE_SET, gst_time, \
                gst.SEEK_TYPE_NONE, 0)
            self.bin.send_event(event)
            self.dispatch_signame('player.status')

    def _update_metadata(self, tags):
        if not self._media_file or self._media_file["type"] != "webradio":
            return

        for k in tags.keys():
            value = str(tags[k]).strip()
            if not value: continue
            if k in ("emphasis", "mode", "layer"):
                continue
            elif isinstance(value, basestring):
                if k in ("title", "album", "artist"):
                    name = "song-" + k
                    if name not in self._media_file.keys() or\
                                   self._media_file[name] != value:
                        self._media_file[name] = value
        self.dispatch_signame('player.current')

    def get_state(self):
        gst_state = self.__get_gst_state()
        if gst_state == gst.STATE_PLAYING:
            return PLAYER_PLAY
        elif gst_state == gst.STATE_PAUSED:
            return PLAYER_PAUSE
        else:
            return PLAYER_STOP

    def __get_gst_state(self):
        changestatus,state,_state = self.bin.get_state()
        return state

    def is_supported_uri(self,uri_type):
        if uri_type == "dvd": return False
        return gst.element_make_from_uri(gst.URI_SRC,uri_type+"://", '') \
                    is not None

    def is_supported_format(self,format):
        formats = {
            ".mp3": ("mad",),
            ".mp2": ("mad",),
            ".ogg": ("ogg", "vorbis"),
            ".mp4": ("faad",),
            ".flac": ("flac",),
            }

        if format in formats.keys():
            for plugin in formats[format]:
                if gst.registry_get_default().find_plugin(plugin) == None:
                    return False
            return True

        return False

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import urllib
from zope.interface import implements
from deejayd.xmlobject import ET
from deejayd.plugins import PluginError, IWebradioPlugin

# Without arg return genre list
# with genre= return station list
SHOUTCAST_URL_QUERY = "http://www.shoutcast.com/sbin/newxml.phtml"

# Use id= to select stream
SHOUTCAST_URL_STREAM = "http://www.shoutcast.com"
SHOUTCAST_TUNEIN_BASE = "/sbin/tunein-station.pls"

class ShoutcastPlugin(object):
    implements(IWebradioPlugin)
    NAME="shoutcast"
    HAS_CATEGORIE = True

    def __init__(self):
        self.__tunein = {}
        self.__categories = None
        self.__streams = {}

    def get_xml_page(self, url):
        try:
            page_handle = urllib.urlopen(url)
            xml_page = page_handle.read()
        except:
            raise PluginError(_("Unable to connect to shoutcast website"))

        # try to parse result
        try:
            root = ET.fromstring(xml_page)
        except ET.XMLSyntaxError:
            raise PluginError(_("Unable to parse shoutcast website page: %s")\
                              %xml_page)
        except:
            raise PluginError(_("Unable to read result from shoutcast website"))
        finally:
            page_handle.close()

        return root

    def get_categories(self):
        if self.__categories is None:
            root = self.get_xml_page(SHOUTCAST_URL_QUERY)
            if root is None:
                return []
            self.__categories = []
            for genre_elt in root.getchildren():
                if "name" in genre_elt.attrib:
                    self.__categories.append(genre_elt.attrib["name"])

        return self.__categories

    def get_streams(self, categorie):
        if categorie == "":
            return []

        if categorie not in self.__streams.keys():
            # get streams for this cat
            root = self.get_xml_page(SHOUTCAST_URL_QUERY+"?genre="+categorie)
            if root is None:
                return []

            self.__streams[categorie] = []
            for station in root.getchildren():
                if station.tag == "tunein":
                    self.__tunein[categorie] = station.attrib["base"]
                elif station.tag == "station":
                    tunein = categorie in self.__tunein.keys() and \
                        self.__tunein[categorie] or SHOUTCAST_TUNEIN_BASE
                    self.__streams[categorie].append({
                        "title": station.attrib["name"],
                        "type": "webradio",
                        "url": SHOUTCAST_URL_STREAM + tunein + "?id="\
                             + station.attrib["id"],
                        "url-type": "pls",
                        "uri": "",
                        })

        return self.__streams[categorie]

# vim: ts=4 sw=4 expandtab


# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Deejayd JSON-RPC protocol testing"""

import re
from testdeejayd import TestCaseWithAudioAndVideoData, TestCaseWithData
from deejayd.mediafilters import *
from deejayd.interfaces import DeejaydSignal
from deejayd.net.client import _DeejayDaemon,\
        DeejaydAnswer, DeejaydKeyValue, DeejaydList,\
        DeejaydFileList, DeejaydMediaList,\
        DeejaydStaticPlaylist, DeejaydError
from deejayd.rpc.jsonbuilders import JSONRPCRequest, JSONRPCResponse,\
                                     Get_json_filter, DeejaydJSONSignal

def trim_json(json):
    return re.sub('(\s{2,})|(\\n)',' ', json)

class TestJSONRPCBuilders(TestCaseWithData):

    def test_request_builder(self):
        """ test JSON-RPC request building """
        cmd = JSONRPCRequest("m_name", ["args1", 2])
        expected_answer = """{"params": ["args1", 2],
                "method": "m_name",
                "id": %d}""" % cmd.get_id()
        self.assertEqual(trim_json(cmd.to_json()), trim_json(expected_answer))

    def test_response_builder(self):
        """ test JSON-RPC response building """
        expected_answer = """{"id": 22,
                "result": {"answer": ["r1", "r2"],
                "type": "type1"}, "error": null}"""
        ans = JSONRPCResponse({"type": "type1", "answer": ["r1", "r2"]}, 22)
        self.assertEqual(trim_json(ans.to_json()), trim_json(expected_answer))

    def test_filter_builder(self):
        """ test JSON-RPC mediafilter building """
        filter = And(Equals("artist", "artist_name"),\
                Or(Contains("genre", "Rock"), Higher("Rating", "4")))
        expected_answer = """{"type": "complex",
            "id": "and",
            "value": [{"type": "basic",
                       "id": "equals",
                       "value": {"pattern": "artist_name", "tag": "artist"}},
                      {"type": "complex",
                       "id": "or",
                       "value": [{"type": "basic",
                                  "id": "contains",
                                  "value": {"pattern": "Rock",
                                            "tag": "genre"}},
                                           {"type": "basic",
                                            "id": "higher",
                                            "value": {"pattern": "4",
                                                      "tag": "Rating"}}]}]}"""
        f = Get_json_filter(filter)
        self.assertEqual(trim_json(f.to_json()), trim_json(expected_answer))

    def test_signal_builder(self):
        """ test JSON-RPC Signal building """
        signal = DeejaydSignal("signal_name", {"attr1": "value1", "attr2": 22})
        expected_answer = """{"answer":
            {"name": "signal_name",
             "attrs": {"attr2": 22, "attr1": "value1"}},
            "type": "signal"}"""
        s = DeejaydJSONSignal(signal)
        self.assertEqual(trim_json(s.to_json()), trim_json(expected_answer))


class TestAnswerParser(TestCaseWithData):
    """Test the Deejayd client library answer parser"""

    def setUp(self):
        super(TestAnswerParser, self).setUp()
        self.deejayd = _DeejayDaemon()

    def parse_answer(self, str):
        self.deejayd._build_answer(str)

    def test_answer_parser_wrongformat(self):
        """Test the client library parsing a wrong format answer"""
        wrong_answer = """{"id": 1,
                "result": {"answer": true,
                "type": "ack"}}"""
        ans = DeejaydAnswer()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.assertRaises(DeejaydError, self.parse_answer, wrong_answer)

    def test_answer_parser_wrongid(self):
        """Test the client library parsing an answer with wrong id"""
        wrongid_answer = """{"id": 1,
                "error": null,
                "result": {"answer": true,
                "type": "ack"}}"""
        ans = DeejaydAnswer()
        ans.set_id(2)
        self.deejayd.expected_answers_queue.put(ans)
        self.assertRaises(DeejaydError, self.parse_answer, wrongid_answer)

    def test_answer_parser_ack(self):
        """Test the client library parsing an ack answer"""
        ack_answer = """{"id": 1,
                "result": {"answer": true,
                "type": "ack"}, "error": null}"""
        ans = DeejaydAnswer()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.parse_answer(ack_answer)

        self.failUnless(ans.get_contents())

    def test_answer_parser_keyvalue(self):
        """Test the client library parsing a key value answer"""
        how_much = 10
        orig_key_value = {}
        for count in range(how_much):
            key = self.testdata.getRandomString()
            value =  self.testdata.getRandomString()
            orig_key_value[key] = value
        key_value_answer = """{"id": 1,
                "error": null,
                "result": {"type": "dict",
                           "answer": {%s}}}""" \
            % ",".join([""" "%s": "%s" """ % (k,v) \
                            for k,v in orig_key_value.items()])

        ans = DeejaydKeyValue()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.parse_answer(key_value_answer)

        retrieved_key_values = ans.get_contents()
        for key in orig_key_value.keys():
            self.assertEqual(orig_key_value[key], retrieved_key_values[key])

    def test_answer_parser_list(self):
        """Test the client library parsing a list answer."""
        how_much = 10
        orig_list = []
        for count in range(how_much):
            item = self.testdata.getRandomString()
            orig_list.append(item)
        list_answer = """{"id": 1,
                "error": null,
                "result": {"type": "list",
                           "answer": [%s]}}""" \
            % ",".join([""" "%s" """ % v for v in orig_list])

        ans = DeejaydList()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.parse_answer(list_answer)
        retrieved_list = ans.get_contents()

        for item in orig_list:
            self.failUnless(item in retrieved_list)

    def test_answer_parser_fileanddir_list(self):
        """Test the client library parsing a file/dir list answer"""
        def dict_to_json(input):
            return """{%s}""" % ",".join([""" "%s": "%s" """ % (k,v)\
                    for k,v in input.items()])

        how_much = self.testdata.getRandomInt(50)
        orig_files, orig_dirs = [], []
        for count in range(how_much):
            if self.testdata.getRandomElement(['file','directory']) == 'file':
                file = {}
                file['id'] = count
                how_much_parms = self.testdata.getRandomInt()
                for parm_count in range(how_much_parms):
                    name = self.testdata.getRandomString()
                    value =  self.testdata.getRandomString()
                    file[name] = value

                orig_files.append(file)
            else:
                dirname = self.testdata.getRandomString()
                orig_dirs.append(dirname)
        fandd_list_answer = """{"id": 1,
                "error": null,
                "result": {"type": "fileAndDirList",
                           "answer": {"type": "song",
                                      "directories": [%(dirs)s],
                                      "files": [%(files)s],
                                      "root": ""}}}""" \
            % {
                "dirs": ",".join([""" "%s" """ % v for v in orig_dirs]),
                "files": ",".join([""" %s """ % dict_to_json(v)\
                                              for v in orig_files]),
            }

        ans = DeejaydFileList()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.parse_answer(fandd_list_answer)

        for file in orig_files:
            # Find corresponding file in retrieved files
            found = False
            for retrieved_file in ans.get_files():
                if file['id'] == int(retrieved_file['id']):
                    found = True
                    break

            for key in file.keys():
                self.failUnless(key in retrieved_file.keys())
                self.assertEqual(unicode(file[key]), retrieved_file[key])

        for dir in orig_dirs:
            self.failUnless(unicode(dir) in ans.get_directories())

    def test_answer_parser_medialist(self):
        """Test the client library parsing a medialist answer"""
        def dict_to_json(input):
            return """{%s}""" % ",".join([""" "%s": "%s" """ % (k,v)\
                    for k,v in input.items()])

        how_much = self.testdata.getRandomInt(50)
        orig_medias = []
        for count in range(how_much):
            media = {}
            media['id'] = count
            how_much_parms = self.testdata.getRandomInt()
            for parm_count in range(how_much_parms):
                name = self.testdata.getRandomString()
                value =  self.testdata.getRandomString()
                media[name] = value
            orig_medias.append(media)

        filter = And(Equals("artist", "artist_name"),\
                Or(Contains("genre", "Rock"), Higher("Rating", "4")))
        json_filter = Get_json_filter(filter)

        medialist_answer = """{"id": 1,
                "error": null,
                "result": {"type": "mediaList",
                           "answer": {"media_type": "song",
                                      "medias": [%(medias)s],
                                      "filter": %(filter)s}}}""" \
            % {
                "filter": json_filter.to_json(),
                "medias": ",".join([""" %s """ % dict_to_json(v)\
                                              for v in orig_medias]),
            }

        ans = DeejaydMediaList()
        ans.set_id(1)
        self.deejayd.expected_answers_queue.put(ans)
        self.parse_answer(medialist_answer)

        for media in orig_medias:
            # Find corresponding media
            found = False
            for retrieved_media in ans.get_medias():
                if media['id'] == int(retrieved_media['id']):
                    found = True
                    break

            for key in media.keys():
                self.failUnless(key in retrieved_media.keys())
                self.assertEqual(unicode(media[key]), retrieved_media[key])

        self.failUnless(ans.is_magic())
        self.failUnless(filter.equals(ans.get_filter()))

    def test_answer_parser_signal(self):
        """Parse a signal message"""
        sig_name = self.testdata.getRandomElement(DeejaydSignal.SIGNALS)
        raw_sig = """{"answer":
            {"name": "%s",
             "attrs": {"attr2": 22, "attr1": "value1"}},
            "type": "signal"}""" % sig_name
        signal_answer = """{"id": null,
                "error": null,
                "result": %s}""" % trim_json(raw_sig)

        # We use a list here as a workaround of python nested scopes
        # limitation : 'sig = None' and then in sig_received() 'sig = signal'
        # would not work.
        sig = []
        def sig_received(signal):
            sig.append(signal)
        self.deejayd.subscribe(sig_name, sig_received)

        self.parse_answer(signal_answer)
        self.assertEqual(sig.pop().get_name(), sig_name)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import random

class Order(object):

    # Not called directly, but the default implementation of
    # next_explicit and next_implicit both just call this.
    def next(self, medialist, current):
        raise NotImplementedError

    # Not called directly, but the default implementation of
    # previous_explicit calls this. Right now there is no such thing
    # as previous_implicit.
    def previous(self, medialist, current):
        raise NotImplementedError

    # Not called directly, but the default implementation of
    # set_explicit calls this. Right now there is no such thing as
    # set_implicit.
    def set(self, medialist, media_id):
        return medialist.get_item(media_id)

    # Called when the user presses a "Next" button.
    def next_explicit(self, medialist, current):
        return self.next(medialist, current)

    # Called when a media ends passively, e.g. it plays through.
    def next_implicit(self, medialist, current):
        return self.next(medialist, current)

    # Called when the user presses a "Previous" button.
    def previous_explicit(self, medialist, current):
        return self.previous(medialist, current)

    # Called when the user manually selects a media
    # If desired the play order can override that
    def set_explicit(self, medialist, media_id):
        return self.set(medialist, media_id)

    def reset(self, medialist):
        pass

class OrderInOrder(Order):
    name = "inorder"

    def next(self, medialist, current):
        if current is None:
            return medialist.get_item_first()
        else:
            next = medialist.next(current)
            if next is None and medialist.repeat:
                next = medialist.get_item_first()
            return next

    def previous(self, medialist, current):
        if len(medialist) == 0:
            return None
        elif current is None:
            return medialist.get_item_last()
        else:
            previous = medialist.previous(current)
            if previous is None and medialist.repeat:
                previous = medialist.get_item_last()
            return previous

class OrderRemembered(Order):
    # Shared class for all the shuffle modes that keep a memory
    # of their previously played medias.

    def __init__(self):
        self._played = []

    def next(self, medialist, current):
        if current is not None:
            self._played.append(current["id"])

    def previous(self, medialist, current):
        try: id = self._played.pop()
        except IndexError: return None
        else: return medialist.get_item(id)

    def set(self, medialist, media_id):
        self._played.append(media_id)
        return medialist.get_item(media_id)

    def reset(self, medialist):
        self._played = []

class OrderShuffle(OrderRemembered):
    name = "random"

    def next(self, medialist, current):
        super(OrderShuffle, self).next(medialist, current)
        played = set(self._played)
        medias = set([m["id"] for m in medialist.get()])
        remaining = medias.difference(played)

        if remaining:
            return medialist.get_item(random.choice(list(remaining)))
        elif medialist.repeat and not medialist.is_empty():
            del(self._played[:])
            return medialist.get_item(random.choice(medias))
        else:
            del(self._played[:])
            return None

class OrderWeighted(OrderRemembered):
    name = "random-weighted"

    def next(self, medialist, current):
        super(OrderWeighted, self).next(medialist, current)
        played = set(self._played)
        remaining = [(m["id"],int(m["rating"])) \
                        for m in medialist.get() if m["id"] not in played]

        max_score = sum([r for (id,r) in remaining])
        choice = int(random.random() * max_score)
        current = 0
        for id, rating in remaining:
            current += rating
            if current >= choice:
                return medialist.get_item(id)

        if medialist.repeat and not medialist.is_empty():
            del(self._played[:])
            return medialist.get_item_first()
        else:
            del(self._played[:])
            return None

class OrderOneMedia(OrderInOrder):
    name = "onemedia"

    def next_implicit(self, medialist, current):
        if medialist.repeat:
            return current
        else:
            return None

orders = {
    "onemedia": OrderOneMedia(),
    "inorder": OrderInOrder(),
    "random": OrderShuffle(),
    "random-weighted": OrderWeighted(),
    }

# vim: ts=4 sw=4 expandtab



#!/usr/bin/env python

# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import glob,os
from distutils.dep_util import newer
from distutils.spawn import spawn


def update_translation(po_dir, po_package, po):
    # Update runtime translations
    os.chdir(po_dir)
    spawn(["intltool-update", "--dist", "--gettext-package", po_package,
           os.path.basename(po[:-3])])


def update_template(po_dir, po_package):
    os.chdir(po_dir)
    # We force here the python language for gettext strings extraction because
    # xgettext (which is called by intltool-update) reverts to C for
    # templates, as they do not have a .py extension, and this does not work
    # and translatable strings are not extracted.
    os.environ['XGETTEXT_ARGS'] = "-L Python"
    spawn(["intltool-update", "--pot", "--gettext-package", po_package])


def update_po():
    po_package = "deejayd"
    po_dir = os.path.abspath("po")

    pot_file = os.path.join(po_dir, po_package + ".pot")
    po_files = glob.glob(os.path.join(po_dir, "*.po"))
    infilename = os.path.join(po_dir, "POTFILES.in")
    infiles = file(infilename).read().splitlines()

    oldpath = os.getcwd()

    need_tpl_update = False
    for filename in infiles:
        if newer(filename, pot_file):
            need_tpl_update = True

    if need_tpl_update:
        update_template(po_dir, po_package)
        for po in po_files:
            update_translation(po_dir, po_package, po)

    os.chdir(oldpath)

if __name__ == "__main__":
    update_po()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os,shutil
from twisted.web import static,server
from twisted.web.resource import Resource

from deejayd.interfaces import DeejaydError
from deejayd.ui import log

# jsonrpc import
from deejayd.webui.jsonrpc import JSONRPC
from deejayd.rpc.protocol import build_protocol, set_web_subhandler

# xul parts
from deejayd.webui.xul import build as xul_build
# mobile parts
from deejayd.webui import mobile


class DeejaydWebError(DeejaydError): pass

class DeejaydMainHandler(Resource):

    def getChild(self, name, request):
        if name == '': return self
        return Resource.getChild(self,name,request)

    def render_GET(self, request):
        user_agent = request.getHeader("user-agent");
        root = request.prepath[-1] != '' and request.path + '/' or request.path
        if user_agent.lower().find("mobile") != -1:
            # redirect to specific mobile interface
            request.redirect(root + 'm/')
            return 'redirected'
        else: # default xul interface
            request.redirect(root + 'xul/')
            return 'redirected'


class DeejaydXulHandler(Resource):

    def __init__(self, config):
        Resource.__init__(self)
        self.__config = config

    def getChild(self, name, request):
        if name == '': return self
        return Resource.getChild(self,name,request)

    def render_GET(self, request):
        request.setHeader("Content-Type", "application/vnd.mozilla.xul+xml")
        return xul_build(self.__config)


class DeejaydMobileHandler(Resource):

    def __init__(self, deejayd, config):
        Resource.__init__(self)
        self.__deejayd = deejayd
        self.__config = config

    def getChild(self, name, request):
        if name == '': return self
        return Resource.getChild(self,name,request)

    def render_GET(self, request):
        # Trailing slash is required for js script paths in the mobile webui,
        # therefore we need to add it if it is missing, by issuing a redirect
        # to the web browser.
        if request.prepath[-1] != '':
            request.redirect(request.path + '/')
            return 'redirected'

        user_agent = request.getHeader("user-agent");
        return mobile.build_template(self.__deejayd, user_agent)


class SiteWithCustomLogging(server.Site):

    def _openLogFile(self, path):
        self.log_file = log.LogFile(path, False)
        self.log_file.set_reopen_signal(callback=self.__reopen_cb)
        return self.log_file.fd

    def __reopen_cb(self, signal, frame):
        self.log_file.reopen()
        # Change the logfile fd from HTTPFactory internals.
        self.logFile = self.log_file.fd


def init(deejayd_core, config, webui_logfile, htdocs_dir):
    # create tmp directory
    tmp_dir = config.get("webui","tmp_dir")
    if os.path.isdir(tmp_dir):
        try: shutil.rmtree(tmp_dir)
        except (IOError, OSError):
            raise DeejaydWebError(_("Unable to remove tmp directory %s") % \
                    tmp_dir)
    try: os.mkdir(tmp_dir)
    except IOError:
        raise DeejaydWebError(_("Unable to create tmp directory %s") % tmp_dir)

    if not os.path.isdir(htdocs_dir):
        raise DeejaydWebError(_("Htdocs directory %s does not exists") % \
                htdocs_dir)

    # main handler
    main_handler = DeejaydMainHandler()
    # json-rpc handler
    rpc_handler = JSONRPC(deejayd_core)
    rpc_handler = build_protocol(deejayd_core, rpc_handler)
    rpc_handler = set_web_subhandler(deejayd_core, tmp_dir, rpc_handler)
    main_handler.putChild("rpc",rpc_handler)
    # statics url
    main_handler.putChild("tmp",static.File(tmp_dir))
    main_handler.putChild("static",static.File(htdocs_dir))

    # xul part
    xul_handler = DeejaydXulHandler(config)
    main_handler.putChild("xul", xul_handler)

    # mobile part
    mobile_handler = DeejaydMobileHandler(deejayd_core, config)
    main_handler.putChild("m", mobile_handler)

    return SiteWithCustomLogging(main_handler, logPath=webui_logfile)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from ConfigParser import NoOptionError
from deejayd.ui import log
from deejayd.mediadb import library

try:
    from deejayd.mediadb import inotify
except ImportError:
    inotify = False


def init(db, player, config):
    audio_library,video_library,lib_watcher = None, None, None
    fc = config.get("mediadb","filesystem_charset")

    audio_dir = config.get("mediadb","music_directory")
    try: audio_library = library.AudioLibrary(db, player, audio_dir, fc)
    except library.NotFoundException,msg:
        log.err(_("Unable to init audio library : %s") % msg, fatal = True)

    activated_sources = config.getlist('general', "activated_modes")
    if "video" in activated_sources:
        video_dir = config.get('mediadb', 'video_directory')
        try: video_library = library.VideoLibrary(db,player,video_dir,fc)
        except library.NotFoundException,msg:
            log.err(_("Unable to init video library : %s") % msg, fatal=True)

    if inotify:
        lib_watcher = inotify.get_watcher(db, audio_library, video_library)
    else: log.info(_("Inotify support disabled"))

    return audio_library,video_library,lib_watcher

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import random
from deejayd.sources._base import _BaseAudioLibSource, SourceError

class QueueSource(_BaseAudioLibSource):
    base_medialist = "__djqueue__"
    name = "queue"
    source_signal = 'queue.update'
    available_playorder = ("inorder", "random")
    has_repeat = False

    def go_to(self, nb, type = "id"):
        self._current = super(QueueSource, self).go_to(nb, type)
        if self._current != None:
            self._media_list.delete([self._current["id"],])
            self.dispatch_signame(self.source_signal)
        return self._current

    def next(self, explicit = True):
        self._current = None
        super(QueueSource, self).next(explicit)
        if self._current != None:
            self._media_list.delete([self._current["id"],])
            self.dispatch_signame(self.source_signal)
        return self._current

    def previous(self,rd,rpt):
        # Have to be never called
        raise NotImplementedError

    def reset(self):
        self._current = None

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys, traceback
from twisted.application import service, internet
from twisted.internet import protocol, reactor
from twisted.internet.error import ConnectionDone
from twisted.protocols.basic import LineReceiver

from deejayd import __version__
from deejayd.interfaces import DeejaydSignal
from deejayd.mediafilters import *
from deejayd.ui import log
from deejayd.utils import str_encode
from deejayd.rpc import Fault, DEEJAYD_PROTOCOL_VERSION
from deejayd.rpc.jsonparsers import loads_request
from deejayd.rpc.jsonbuilders import JSONRPCResponse, DeejaydJSONSignal
from deejayd.rpc import protocol as deejayd_protocol

class DeejaydProtocol(LineReceiver, deejayd_protocol.DeejaydTcpJSONRPC):
    NOT_FOUND = 8001
    FAILURE = 8002
    delimiter = 'ENDJSON\n'

    def __init__(self, deejayd_core, protocol_manager):
        super(DeejaydProtocol, self).__init__()
        self.MAX_LENGTH = 40960
        self.subHandlers = {}
        self.deejayd_core = deejayd_core
        self.manager = protocol_manager

    def connectionMade(self):
        self.send_buffer("OK DEEJAYD %s protocol %d\n" %
                         (__version__, DEEJAYD_PROTOCOL_VERSION, ))

    def connectionLost(self, reason=ConnectionDone):
        self.manager.close_signals(self)

    def lineReceived(self, line):
        line = line.strip("\r")
        # DEBUG Informations
        log.debug(line)

        need_to_close = False
        try:
            parsed = loads_request(line)
            args, functionPath = parsed['params'], parsed["method"]
            function = self._getFunction(functionPath)
            if parsed["method"] == "close":
                # close the connection after send answer
                need_to_close = True
        except Fault, f:
            try: id = parsed["id"]
            except:
                id = None
            ans = JSONRPCResponse(f, id)
        else:
            try: result = function(*args)
            except Exception, ex:
                if not isinstance(ex, Fault):
                    log.err(str_encode(traceback.format_exc()))
                    result = Fault(self.FAILURE, _("error, see deejayd log"))
                else:
                    result = ex
            ans = JSONRPCResponse(result, parsed["id"])

        self.send_buffer(ans.to_json()+self.delimiter)
        if need_to_close:
            self.transport.loseConnection()

    def send_buffer(self, buf):
        if isinstance(buf, unicode):
            buf = buf.encode("utf-8")
        self.transport.write(buf)
        log.debug(buf)

    def lineLengthExceeded(self, line):
        log.err(_("Request too long, close the connection"))
        self.transport.loseConnection()

    def set_signaled(self, signal_name):
        self.manager.set_signaled(self, signal_name)

    def set_not_signaled(self, signal_name):
        self.manager.set_not_signaled(self, signal_name)


class DeejaydFactory(protocol.ServerFactory):
    protocol = DeejaydProtocol
    obj_supplied = False

    def __init__(self, deejayd_core):
        self.deejayd_core = deejayd_core
        self.signaled_clients = dict([(signame, []) for signame\
                                                    in DeejaydSignal.SIGNALS])
        self.core_sub_ids = {}

    def startFactory(self):
        log.info(_("Net Protocol activated"))

    def buildProtocol(self, addr):
        p = self.protocol(self.deejayd_core, self)
        p = deejayd_protocol.build_protocol(self.deejayd_core, p)
        # set specific signal subhandler
        p = deejayd_protocol.set_signal_subhandler(self.deejayd_core, p)
        p.factory = self
        return p

    def clientConnectionLost(self, connector, reason):
        for signal_name in DeejaydSignal.SIGNALS:
            self.set_not_signaled(connector, signal_name)

    def set_signaled(self, connector, signal_name):
        client_list =  self.signaled_clients[signal_name]
        if len(client_list) < 1:
            # First subscription for this signal, so subscribe
            sub_id = self.deejayd_core.subscribe(signal_name,
                                                 self.sig_bcast_to_clients)
            self.core_sub_ids[signal_name] = sub_id

        self.signaled_clients[signal_name].append(connector)

    def set_not_signaled(self, connector, signal_name):
        client_list =  self.signaled_clients[signal_name]
        if connector in client_list:
            client_list.remove(connector)
        if len(client_list) < 1:
            # No more clients for this signal, we can unsubscribe
            self.deejayd_core.unsubscribe(self.core_sub_ids[signal_name])

    def close_signals(self, connector):
        for signal_name in self.signaled_clients.keys():
            if len(self.signaled_clients[signal_name]) > 0:
                self.set_not_signaled(connector, signal_name)

    def sig_bcast_to_clients(self, signal):
        interested_clients = self.signaled_clients[signal.get_name()]
        if len(interested_clients) > 0:
            j_sig = DeejaydJSONSignal(signal)
            ans = JSONRPCResponse(j_sig.dump(), None).to_json()
            for client in interested_clients:
                # http://twistedmatrix.com/pipermail/twisted-python/2007-August/015905.html
                # says : "Don't call reactor methods from any thread except the
                # one which is running the reactor.  This will have
                # unpredictable results and generally be broken."
                # This is the "why" of this weird call instead of a simple
                # client.send_buffer(xml_sig.to_xml()).
                reactor.callFromThread(client.send_buffer, ans+client.delimiter)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from deejayd.mediadb.formats._base import _AudioFile

extensions = [".mp3",".mp2",".aac"]
try: from mutagen.mp3 import MP3
except ImportError:
    extensions = []


class Mp3File(_AudioFile):
    IDS = { "TIT2": "title",
            "TPE1": "artist",
            "TALB": "album",
            }
    replaygain_process = False

    def parse(self, file):
        infos = _AudioFile.parse(self, file)
        mp3_info = MP3(file)

        infos.update([
            ("title", ""),
            ("artist", ""),
            ("various_artist", ""),
            ("album", ""),
            ("tracknumber", ""),
            ("discnumber", ""),
            ("date", ""),
            ("genre", ""),
            ("replaygain_track_gain", ""),
            ("replaygain_track_peak", ""),
            ("bitrate", int(mp3_info.info.bitrate)),
            ("length", int(mp3_info.info.length)),
            ])

        tag = mp3_info.tags
        if not tag:
            infos["title"] = os.path.split(file)[1]
            return infos

        for frame in tag.values():
            if frame.FrameID == "TXXX":
                if frame.desc in ("replaygain_track_peak",\
                                  "replaygain_track_gain"):
                    # Some versions of Foobar2000 write broken Replay Gain
                    # tags in this format.
                    infos[frame.desc] = frame.text[0]
                    self.replaygain_process = True
                else: continue
            elif frame.FrameID == "RVA2": # replaygain
                self.__process_rg(frame, infos)
                continue
            elif frame.FrameID == "TCON": # genre
                infos["genre"] = frame.genres[0]
                continue
            elif frame.FrameID == "TDRC": # date
                list = [stamp.text for stamp in frame.text]
                infos["date"] = list[0]
                continue
            elif frame.FrameID == "TRCK": # tracknumber
                infos["tracknumber"] = self._format_number(frame.text[0])
            elif frame.FrameID == "TPOS": # discnumber
                infos["discnumber"] = self._format_number(frame.text[0])
            elif frame.FrameID in self.IDS.keys():
                infos[self.IDS[frame.FrameID]] = frame.text[0]
            elif frame.FrameID == "APIC": # picture
                if frame.type == 3: # album front cover
                    infos["cover"] = {"data": frame.data, "mime": frame.mime}
            else: continue

        infos["various_artist"] = infos["artist"]
        return infos

    def __process_rg(self, frame, infos):
        if frame.channel == 1:
            if frame.desc == "album": return # not supported
            elif frame.desc == "track" or not self.replaygain_process:
                infos["replaygain_track_gain"] = "%+f dB" % frame.gain
                infos["replaygain_track_peak"] = str(frame.peak)
                self.replaygain_process = True

object = Mp3File

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# This work is based on the perl X11::FullScreen API.
# http://search.cpan.org/~stephen/X11-FullScreen-0.03/


import ctypes
import _libX11 as libX11
import _libXext as libXext


class X11Error(Exception): pass


class X11Event:
    NoEventMask              = 0L
    KeyPressMask             = (1L<<0)
    KeyReleaseMask           = (1L<<1)
    ButtonPressMask          = (1L<<2)
    ButtonReleaseMask        = (1L<<3)
    EnterWindowMask          = (1L<<4)
    LeaveWindowMask          = (1L<<5)
    PointerMotionMask        = (1L<<6)
    PointerMotionHintMask    = (1L<<7)
    Button1MotionMask        = (1L<<8)
    Button2MotionMask        = (1L<<9)
    Button3MotionMask        = (1L<<10)
    Button4MotionMask        = (1L<<11)
    Button5MotionMask        = (1L<<12)
    ButtonMotionMask         = (1L<<13)
    KeymapStateMask          = (1L<<14)
    ExposureMask             = (1L<<15)
    VisibilityChangeMask     = (1L<<16)
    StructureNotifyMask      = (1L<<17)
    ResizeRedirectMask       = (1L<<18)
    SubstructureNotifyMask   = (1L<<19)
    SubstructureRedirectMask = (1L<<20)
    FocusChangeMask          = (1L<<21)
    PropertyChangeMask       = (1L<<22)
    ColormapChangeMask       = (1L<<23)
    OwnerGrabButtonMask      = (1L<<24)


class X11Window(object):

    def __init__(self, display, width=None, height=None, fullscreen=False):
        self.__display = display
        self.__display_p = self.__display.display_p()

        libX11.XLockDisplay(self.__display_p)

        if fullscreen:
            self.__always_fulscreen = True
            self.__fullscreen = True
            self.width = self.__display.get_width()
            self.height = self.__display.get_height()
        elif width and height:
            self.__always_fulscreen = False
            self.__fullscreen = False
            self.width = width
            self.height = height
        elif not width and not height and not fullscreen:
            raise X11Error('A window is either fullscreen or has dimensions.')

        self.__window_p = libX11.XCreateSimpleWindow(self.__display_p,
                   libX11.XDefaultRootWindow(self.__display_p),
                   0, 0, self.width, self.height, 0, 0, 0)

        self.__hide_cursor()
        if self.__fullscreen:
            self.__remove_decorations()

        libX11.XSelectInput(self.__display_p, self.__window_p,
                            (X11Event.ExposureMask |\
                             X11Event.ButtonPressMask |\
                             X11Event.KeyPressMask |\
                             X11Event.ButtonMotionMask |\
                             X11Event.StructureNotifyMask |\
                             X11Event.PropertyChangeMask |\
                             X11Event.PointerMotionMask))
        libX11.XMapRaised(self.__display_p, self.__window_p)
        libX11.XSync(self.__display_p, False)
        libX11.XUnlockDisplay(self.__display_p)

        self.video_area_info = {}
        self.update_video_area_info()

    def window_p(self):
        return self.__window_p

    def __remove_decorations(self):
        mwmhints = libX11.MWMHints()
        mwmhints.decorations = 0
        mwmhints.flags = libX11.MWMHints.MWM_HINTS_DECORATIONS
        data = ctypes.cast(ctypes.byref(mwmhints), ctypes.c_char_p)
        xa_no_border = libX11.XInternAtom(self.__display_p,
                                          "_MOTIF_WM_HINTS",
                                          False)
        libX11.XChangeProperty(self.__display_p, self.__window_p,
                               xa_no_border, xa_no_border,
                               32, libX11.PropertyModes.PropModeReplace, data,
                               libX11.MWMHints.PROP_MWM_HINTS_ELEMENTS)

    def __ignore_wm(self):
        attr = libX11.XSetWindowAttributes()
        attr.override_redirect = True
        libX11.XChangeWindowAttributes(self.__display_p, self.__window_p,
                                       libX11.CWOverrideRedirect)

    def __hide_cursor(self):
        black = libX11.XColor()
        dummy = libX11.XColor()
        bitmap_struct = ctypes.c_char * 8
        bm_no_data = bitmap_struct('0', '0', '0', '0', '0', '0', '0', '0')

        cmap = libX11.XDefaultColormap(self.__display_p,
                                   self.__display.get_default_screen_number())
        libX11.XAllocNamedColor(self.__display_p, cmap, "black",
                                ctypes.byref(black), ctypes.byref(dummy))

        bm_no = libX11.XCreateBitmapFromData(self.__display_p,
                                             self.__window_p,
                                             bm_no_data, 1, 1)
        no_ptr = libX11.XCreatePixmapCursor(self.__display_p,
                                            bm_no, bm_no,
                                            ctypes.byref(black),
                                            ctypes.byref(black),
                                            0, 0)
        libX11.XDefineCursor(self.__display_p, self.__window_p, no_ptr)

        libX11.XFreeCursor(self.__display_p, no_ptr)

        if bm_no != None:
            libX11.XFreePixmap(self.__display_p, bm_no)

        pixel = ctypes.c_ulong(black.pixel)
        libX11.XFreeColors(self.__display_p, cmap, ctypes.byref(pixel), 1, 0)

    def get_geometry(self):
        root = ctypes.c_ulong()
        x = ctypes.c_int()
        y = ctypes.c_int()
        width =ctypes. c_uint()
        height = ctypes.c_uint()
        border_width = ctypes.c_uint()
        depth = ctypes.c_uint()

        libX11.XGetGeometry(self.__display_p, self.__window_p,
                            ctypes.byref(root),
                            ctypes.byref(x), ctypes.byref(y),
                            ctypes.byref(width), ctypes.byref(height),
                            ctypes.byref(border_width), ctypes.byref(depth))

        return x.value, y.value, width.value, height.value

    def update_video_area_info(self):
        # FIXME : This should be called when the window is resized.
        x, y, w, h = self.get_geometry()
        self.video_area_info['win_x'] = x
        self.video_area_info['win_y'] = y
        self.video_area_info['width'] = w
        self.video_area_info['height'] = h
        self.video_area_info['aspect'] = self.__display.get_pixel_aspect()

    def set_fullscreen(self, fson=True):
        if fson == self.__fulscreen or self.__always_fulscreen:
            return

        libX11.XLockDisplay(self.__display_p)
        if fson:
            new_width = self.__display.get_width()
            new_height = self.__display.get_height()
            self.__remove_decorations()
            self.__fulscreen = True
        else:
            new_width = self.width
            new_height = self.height
            self.__fulscreen = False
        libX11.XResizeWindow(self.__display_p, self.__window_p,
                             new_width, new_height)
        libX11.XUnlockDisplay(self.__display_p)
        self.update_video_area_info()

    def show(self, do_show=True):
        libX11.XLockDisplay(self.__display_p)
        if do_show:
            libX11.XMapRaised(self.__display_p, self.__window_p)
        else:
            libX11.XUnmapWindow(self.__display_p, self.__window_p)
        libX11.XUnlockDisplay(self.__display_p)

    def close(self):
        libX11.XLockDisplay(self.__display_p)
        libX11.XUnmapWindow(self.__display_p, self.__window_p)
        libX11.XDestroyWindow(self.__display_p, self.__window_p)
        libX11.XUnlockDisplay(self.__display_p)

        self.video_area_info = None

class X11Display(object):

    def __init__(self, id=':0.0'):
        self.id = id

        if not libX11.XInitThreads():
            raise X11Error('Could not init threads')

        self.__display_p = libX11.XOpenDisplay(self.id)
        if not self.__display_p:
            raise X11Error('Could not open display %s' % self.id)

        self.__dpms_orig = self.is_dpms_on()

    def display_p(self):
        return self.__display_p

    def get_default_screen_number(self):
        return libX11.XDefaultScreen(self.__display_p)

    def get_width(self, screen_number=None):
        if not screen_number:
            screen_number = self.get_default_screen_number()
        return libX11.XDisplayWidth(self.__display_p, screen_number)

    def get_height(self, screen_number=None):
        if not screen_number:
            screen_number = self.get_default_screen_number()
        return libX11.XDisplayHeight(self.__display_p, screen_number)

    def get_pixel_aspect(self, screen_number=None):
        if not screen_number:
            screen_number = self.get_default_screen_number()

        res_h = float(self.get_width(screen_number) * 1000 /\
                      libX11.XDisplayWidthMM(self.__display_p, screen_number))
        res_v = float(self.get_height(screen_number) * 1000 /\
                      libX11.XDisplayHeightMM(self.__display_p, screen_number))
        return res_v / res_h

    def is_dpms_on(self):
        dummy = ctypes.c_int()
        if libXext.DPMSQueryExtension(self.__display_p,
                                      ctypes.byref(dummy), ctypes.byref(dummy)):
            return True
        else:
            return False

    def set_dpms(self, activated):
        if self.is_dpms_on():
            if activated:
                libXext.DPMSEnable(self.__display_p)
            else:
                libXext.DPMSDisable(self.__display_p)

    def do_create_window(self, width=None, height=None, fullscreen=False):
        return X11Window(self, width, height, fullscreen)

    def destroy(self):
        self.set_dpms(self.__dpms_orig)
        libX11.XCloseDisplay(self.__display_p)


if __name__ == '__main__':
    import time
    d = X11Display()
    w = d.do_create_window(320, 200)
    time.sleep(5)
    w.close()
    d.destroy()


# vim: ts=4 sw=4 expandtab
